/***************************************************************************
 *   Copyright (C) 2007 by Thomas Keller                                   *
 *   me@thomaskeller.biz                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "Guitone.h"
#include "MainWindow.h"
#include "Monotone.h"
#include "Preferences.h"
#include "Settings.h"

#ifdef Q_WS_MACX
#include "CocoaUtil.h"
#include <Carbon/Carbon.h>
#endif

#include <QMessageBox>
#include <QTimer>
#include <QMutexLocker>
#include <QFileInfo>
#include <QFileDialog>

#ifdef Q_WS_MACX
static OSErr OpenDocumentAE(const AppleEvent * evt, AppleEvent *, long)
{
    AEDescList	docList;
    DescType    retType;
    AEKeyword   keywd;
    FSRef	    fsref;
    long        count, size;
    
    AEGetParamDesc(evt, keyDirectObject, typeAEList, &docList);
    AECountItems(&docList, &count);
    
    OSErr         err;
    
    for (int i = 1; i <= count; i++)
    {
        err = AEGetNthPtr(&docList, i, typeFSRef, &keywd,
                          &retType, (Ptr)&fsref, sizeof(fsref), &size);
        
        if (err != noErr)
        {
            W("Couldn't retrieve FSRef.");
            continue;
        }
        
        QString path = CocoaUtil::FSRefToPath(fsref);
        if (path.isEmpty())
        {
            W("Couldn't convert FSRef to path.");
            continue;
        }
        
        D(QString("Acquired new path: %1").arg(path));
        APP->loadFromPath(path);
    }
    AEDisposeDesc(&docList);
    
    return 0;
}

bool Guitone::macEventFilter(EventHandlerCallRef caller, EventRef event)
{
    //
    // taken from 
    //  http://developer.apple.com/documentation/AppleScript/Conceptual/
    //  AppleEvents/dispatch_aes_aepg/chapter_4_section_3.html
    //
    Boolean     release = false;
    EventRecord eventRecord;
    OSErr       ignoredErr;
 
    // Events of type kEventAppleEvent must be removed from the queue
    //  before being passed to AEProcessAppleEvent.
    if (IsEventInQueue(GetMainEventQueue(), event))
    {
        // RemoveEventFromQueue will release the event, which will
        //  destroy it if we don't retain it first.
        RetainEvent(event);
        release = true;
        RemoveEventFromQueue(GetMainEventQueue(), event);
    }
 
    // Convert the event ref to the type AEProcessAppleEvent expects.
    ConvertEventRefToEventRecord(event, &eventRecord);
    ignoredErr = AEProcessAppleEvent(&eventRecord);
 
    if (release)
        ReleaseEvent(event);
 
    // This Carbon event has been handled, even if no AppleEvent handlers
    //  were installed for the Apple event.
    return noErr;
}
#endif

Guitone::Guitone(int argc, char** argv)
    : QApplication(argc, argv), updateDialog(0)
{
    setQuitOnLastWindowClosed(false);
    
    setOrganizationName("Thomas Keller");
    setOrganizationDomain("thomaskeller.biz");
    setApplicationName("guitone");

#ifdef Q_WS_MACX
    // install apple event handler to open documents
    OSErr err = AEInstallEventHandler(
        kCoreEventClass, kAEOpenDocuments,
        NewAEEventHandlerUPP(OpenDocumentAE), 0, false
    );
    require_noerr(err, error_installing_event_handler);
    return;
    
error_installing_event_handler:
    C("Could not install OpenDocumentAE event handler");
#endif
}

bool Guitone::init()
{
    if (Settings::getBool("CheckForUpdates", true))
    {
#ifdef Q_WS_MACX
        CocoaUtil::initialize();
#else
        updateDialog = new ApplicationUpdate(NULL);
        if (updateDialog->updateAvailable())
        {
            updateDialog->setStayOnTop();
            updateDialog->show();
        }
#endif
    }
    
    QStringList args = arguments();
    bool somethingLoaded = false;
    
    for (int i=1, j=args.size(); i<j; i++)
    {
        somethingLoaded = somethingLoaded || loadFromPath(args.at(i));
    }
    
    // if no command line arguments have been given or
    // they have been invalid or anything else went wrong,
    // try to load the recently used workspace/database
    if (!somethingLoaded)
    {
        QStringList workspaces = Settings::getItemList("RecentWorkspaceList");
        somethingLoaded = workspaces.size() > 0 && loadWorkspace(workspaces.at(0));
    }
    
    if (!somethingLoaded)
    {
        QStringList databases = Settings::getItemList("RecentDatabaseList");
        somethingLoaded = databases.size() > 0 && loadDatabase(databases.at(0));
    }
    
    // if still nothing is loaded, prompt the user to load a workspace
    if (!somethingLoaded)
    {
        QString workspaceDir = 
            QFileDialog::getExistingDirectory(0, tr("Select your workspace..."));
        somethingLoaded = !workspaceDir.isEmpty() && loadWorkspace(workspaceDir);
    }

    // if nothing could be loaded (= no window could be created),
    // let the app close itself
    return somethingLoaded;
}

Guitone::~Guitone()
{
    I(openWindows.size() == 0);
    I(monotoneInstances.size() == 0);
    if (updateDialog) delete updateDialog;
}

bool Guitone::loadFromPath(const QString & path)
{
    QFileInfo fileInfo(path);
    if (!fileInfo.exists())
    {
        W(QString("Non-existant file/folder %1").arg(path));
        return false;
    }
    
    if (fileInfo.isDir())
    {
        return loadWorkspace(path);
    }
    
    return loadDatabase(path);
}

bool Guitone::loadWorkspace(const QString & path)
{
    QMutexLocker locker(&mutex);
    
    MainWindow * wnd = addWindow();
    if (!addMonotoneInstance(wnd) || !wnd->doLoadWorkspace(path))
    {
        removeWindow(wnd);
        return false;
    }
    
    // now that the workspace is loaded, the window should have gotten a
    // reasonable name
    emit windowListChanged();
    
    wnd->show();
    
    return true;
}

bool Guitone::loadDatabase(const QString & path)
{
    QMutexLocker locker(&mutex);
    
    MainWindow * wnd = addWindow();
    if (!addMonotoneInstance(wnd) || !wnd->doLoadDatabase(path))
    {
        removeWindow(wnd);
        return false;
    }
    // now that the database is loaded, the window should have gotten a
    // reasonable name
    emit windowListChanged();
    
    wnd->show();
    
    return true;
}

void Guitone::windowClosed(MainWindow * wnd)
{
    QMutexLocker locker(&mutex);
    
    removeMonotoneInstance(wnd);
    removeWindow(wnd);
    if (openWindows.size() == 0) quit();
}

MainWindow * Guitone::addWindow()
{
    MainWindow * wnd = new MainWindow();
    
    connect(
        wnd, SIGNAL(quitApplication()),
        this, SLOT(quit())
    );
    
    connect(
        wnd, SIGNAL(windowClosed(MainWindow *)),
        this, SLOT(windowClosed(MainWindow *))
    );
    
    connect(
        wnd, SIGNAL(loadWorkspace(const QString &)),
        this, SLOT(loadWorkspace(const QString &))
    );
    
    connect(
        wnd, SIGNAL(loadDatabase(const QString &)),
        this, SLOT(loadDatabase(const QString &))
    );
    
    connect(
        this, SIGNAL(windowListChanged()),
        wnd, SLOT(updateWindowList())
    );
    
    for (int i=0, j=openWindows.size(); i<j; i++)
    {
        MainWindow * other = openWindows.at(i);
        connect(
            wnd, SIGNAL(updatePreviousWorkspacesMenu()),
            other, SLOT(doUpdatePreviousWorkspacesMenu())
        );
        
        connect(
            wnd, SIGNAL(updatePreviousDatabasesMenu()),
            other, SLOT(doUpdatePreviousDatabasesMenu())
        );
        
        connect(
            other, SIGNAL(updatePreviousWorkspacesMenu()),
            wnd, SLOT(doUpdatePreviousWorkspacesMenu())
        );
        
        connect(
            other, SIGNAL(updatePreviousDatabasesMenu()),
            wnd, SLOT(doUpdatePreviousDatabasesMenu())
        );
    }

    openWindows.append(wnd);

    D(QString("Added window %1").arg((qint64)wnd));
    
    return wnd;
}

void Guitone::removeWindow(MainWindow * wnd)
{
    I(wnd);
    
    D(QString("Removing window %1").arg((qint64)wnd));
    
    int i = openWindows.indexOf(wnd);
    if (i == -1) return;
    openWindows.removeAt(i);
    
    disconnect(
        wnd, SIGNAL(quitApplication()),
        this, SLOT(quit())
    );
    
    disconnect(
        wnd, SIGNAL(windowClosed(MainWindow *)),
        this, SLOT(windowClosed(MainWindow *))
    );
    
    disconnect(
        wnd, SIGNAL(loadWorkspace(const QString &)),
        this, SLOT(loadWorkspace(const QString &))
    );
    
    disconnect(
        wnd, SIGNAL(loadDatabase(const QString &)),
        this, SLOT(loadDatabase(const QString &))
    );
    
    disconnect(
        this, SIGNAL(windowListChanged()),
        wnd, SLOT(updateWindowList())
    );
    
    for (int i=0, j=openWindows.size(); i<j; i++)
    {
        MainWindow * other = openWindows.at(i);
        disconnect(
            wnd, SIGNAL(updatePreviousWorkspacesMenu()),
            other, SLOT(doUpdatePreviousWorkspacesMenu())
        );
        
        disconnect(
            wnd, SIGNAL(updatePreviousDatabasesMenu()),
            other, SLOT(doUpdatePreviousDatabasesMenu())
        );
        
        disconnect(
            other, SIGNAL(updatePreviousWorkspacesMenu()),
            wnd, SLOT(doUpdatePreviousWorkspacesMenu())
        );
        
        disconnect(
            other, SIGNAL(updatePreviousDatabasesMenu()),
            wnd, SLOT(doUpdatePreviousDatabasesMenu())
        );
    }
    
    emit windowListChanged();
    delete wnd;
}

const QList<MainWindow*> Guitone::windowList() const
{
    return openWindows;
}

void Guitone::quit()
{
    // do we need to close any windows?
    if (openWindows.size() > 0)
    {
        // let all open windows close themselves to allow them cleaning up
        // stuff and saving settings
        // note that if the last window is closed, quit() is called again,
        // so we return here after closing all windows to avoid recursion
        // problems
        foreach (MainWindow * wnd, openWindows)
        {
            wnd->close();
        }
        return;
    }
    
    Settings::sync();
    QApplication::quit();
}

bool Guitone::addMonotoneInstance(MainWindow * wnd)
{
    I(wnd);
    
    Monotone * mtn = new Monotone(wnd);
    
    // check the current monotone version and prompt the user
    // to enter the correct path if there is a problem
    while (true)
    {
        if (mtn->setMtnBinaryPath(Settings::getMtnBinaryPath())) break;
        
        QMessageBox::critical(
            wnd,
            tr("Error"),
            tr("The path to the monotone binary is either invalid "
             "or points to an older version of monotone. "
             "Guitone requires monotone version %1 "
             "or a monotone with interface version %2 or later.")
            .arg(Monotone::RequiredProgramVersion)
            .arg(Monotone::RequiredInterfaceVersion),
            QMessageBox::Ok, 0, 0
        );
        
        Preferences dialog(wnd);
        
        // if new settings have been entered, try to setup the wrapper
        if (dialog.exec() == QDialog::Accepted) continue;
        
        // the user has rejected the dialog, i.e. made no new settings
        return false;
    }
    
    monotoneInstances.insert(wnd, mtn);
    
    D(QString("Added monotone instance %1 for %2").arg((qint64)mtn).arg((qint64)wnd));
    
    return true;
}

bool Guitone::removeMonotoneInstance(MainWindow * wnd)
{
    I(wnd);
    
    if (!monotoneInstances.contains(wnd)) return false;
   
    Monotone * mtn = monotoneInstances.value(wnd);
    
    D(QString("Removing monotone instance %1 for %2").arg((qint64)mtn).arg((qint64)wnd));
    
    delete mtn;
    monotoneInstances.remove(wnd);
    
    return true;
}

MainWindow * Guitone::findMainWindow(QObject * obj)
{
    do
    {
        if (obj->inherits("MainWindow")) break;
        obj = obj->parent();
    } while (obj);
    
    I(obj);
    MainWindow * wnd = qobject_cast<MainWindow *>(obj);
    return wnd;
}

Monotone * Guitone::getMonotoneInstance(QObject * obj)
{
    MainWindow * wnd = findMainWindow(obj);
    I(wnd);
    I(monotoneInstances.contains(wnd));
    return monotoneInstances.value(wnd);
}

void Guitone::lock()
{
    D("lock called");
    mutex.lock();
}

void Guitone::unlock()
{
    D("unlock called");
    mutex.unlock();
}

