/***************************************************************************
 *   Copyright (C) 2006 by Thomas Keller                                   *
 *   me@thomaskeller.biz                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
 
#include "Attributes.h"
#include "InventoryItem.h"
#include "Monotone.h"
#include "BasicIOParser.h"

#include <QSortFilterProxyModel>

Attributes::Attributes(QObject *parent) 
    : QAbstractItemModel(parent)
{
    attributes = new AttributeList();
    mtnDelegate = new MonotoneDelegate(this);
}

Attributes::~Attributes()
{
    attributes->clear();
    delete attributes;
    delete mtnDelegate;
}

bool Attributes::readAttributes(const QModelIndex & index)
{
    // convert ProxyModel index to real Model index
    QModelIndex sourceIndex = static_cast<const QSortFilterProxyModel*>(index.model())->mapToSource(index);
    // get the Model's item reference out there
    InventoryItem *item = static_cast<InventoryItem*>(sourceIndex.internalPointer());

    // clear current attributes list
    attributes->clear();
    // reset the view
    reset();
    
    if (item->isRootDirectory() || item->isCdUp())
    {
        D("item is pseudo item (root or cdup)");
        return false; 
    }
    
    if (item->hasStatus(InventoryItem::Dropped) || !item->isTracked())
    {
        D("item is not tracked or dropped");
        return false;
    }
    
    QStringList cmd;
    cmd.append("attributes");
    cmd.append(item->getPath());
    
    return mtnDelegate->triggerCommand(cmd);
}

void Attributes::parseOutput()
{
    BasicIOParser parser(AutomateCommand::data);
    I(parser.parse());
    StanzaList list = parser.getStanzas();

    for (int i=0, size = list.size(); i < size; ++i)
    {
        Stanza stanza = list.at(i);
        
        Attribute attr;
        
        bool isItem = false;
        
        for (int j=0, size2 = stanza.size(); j < size2; j++)
        {
            StanzaEntry entry = stanza.at(j);
            
            if (i == 0 && j == 0)
            {
                if (entry.sym != "format_version")
                {
                    W("First entry in first stanza should be used format version.");
                } else
                if (entry.vals.at(0) != "1")
                {
                    W("Format version is unequal to '1' - newer interface version?");
                }
                break;
            }
            
            // we're now only interested in stanzas starting with a "attr" entry
            if (j == 0 && entry.sym != "attr") break;
            
            isItem = true;
            
            if (entry.sym == "attr")
            {
                I(entry.vals.size() == 2);
                attr.key = entry.vals.at(0);
                attr.value = entry.vals.at(1);
                continue;
            }
            
            if (entry.sym == "state")
            {
                I(entry.vals.size() == 1);
                QString state = entry.vals.at(0);
                
                if (state == "added")          attr.state = Attribute::Added;
                else if (state == "dropped")   attr.state = Attribute::Dropped;
                else if (state == "changed")   attr.state = Attribute::Changed;
                else if (state == "unchanged") attr.state = Attribute::Unchanged;
                else 
                    W(QString("Unknown attribute state '%1'.").arg(state));
                
                continue;
            }
            W(QString("Unknown symbol %1.").arg(entry.sym));
        }
        
        // check if we really processed an item entry
        if (!isItem) continue;
        attributes->append(attr);
    }    
    
    // reset the view
    reset();
    
    // signal that we've finished (whoever listens to that)
    emit attributesRead();
}

int Attributes::columnCount(const QModelIndex &parent) const
{
	return 3;
}

QVariant Attributes::data(const QModelIndex &index, int role) const
{
	if (!index.isValid())
	{
		return QVariant();
	}
    
    if (role != Qt::DisplayRole)
	{
        return QVariant();
    }
	
    int row = index.row();
    if (row >= attributes->size()) return QVariant();
    
    Attribute attr = attributes->at(row);
    
    switch (index.column())
    {
        case 0: return QVariant(attr.key);
        case 1: return QVariant(attr.value);
        case 2:
            switch (attr.state)
            {
                case Attribute::Added:     return QVariant(tr("added"));
                case Attribute::Dropped:   return QVariant(tr("dropped"));
                case Attribute::Changed:   return QVariant(tr("changed"));
                case Attribute::Unchanged: return QVariant(tr("unchanged"));
            }
    }
    
    return QVariant();
}

Qt::ItemFlags Attributes::flags(const QModelIndex &index) const
{    
	// TODO: add ItemIsEditable as soon as we can set/drop attributes through
    // the automation interface, implement setData() then as well
	return Qt::ItemIsEnabled | Qt::ItemIsSelectable;
}

QVariant Attributes::headerData(int section, Qt::Orientation orientation, int role) const
{
	if (orientation == Qt::Horizontal && role == Qt::DisplayRole)
	{
        switch (section)
        {
            case 0: return QVariant(tr("Key"));
            case 1: return QVariant(tr("Value"));
            case 2: return QVariant(tr("State"));
        }
    }
    return QVariant();
}

int Attributes::rowCount(const QModelIndex& parent) const
{
	return attributes->size();
}

QModelIndex Attributes::index(int row, int column, const QModelIndex& parent) const
{
    return hasIndex(row, column, parent) ? createIndex(row, column, 0) : QModelIndex();
}

QModelIndex Attributes::parent(const QModelIndex& index) const
{
    return QModelIndex();
}

