/***************************************************************************
 *   Copyright (C) 2006 by Jean-Louis Fuchs                                 *
 *   ganwell@fangorn.ch                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "ChangesetModel.h"

#include <QFont>

ChangesetModel::ChangesetModel(QObject *parent) : QAbstractItemModel(parent)
{
    currentBranch = "";
}

ChangesetModel::~ChangesetModel() {}

void ChangesetModel::receiveRevisions(bool all)
{
    int i = branchMap[currentBranch].certsRead;
    int to = branchMap[currentBranch].revisions.count();
    int c = 0;
    while(i < to && (c < revPerReceive || all))
    {
        QString revision = branchMap[currentBranch].revisions[i];
                
        if(changesetMap[revision] == NULL)
        {
            ChangesetCerts *certs = new ChangesetCerts(currentBranch, revision, i, this);
            certs->readCerts(revision);
            c++;
        }
        i++;
    }
}

void ChangesetModel::setBranch(QString branch)
{
    currentBranch = branch;
    BranchInfo *branchInfo = &branchMap[currentBranch];
    if(branchInfo->revisions.count() == 0)
    {
        branchInfo->certsRead = 0;
        GetBranchLog *glog = new GetBranchLog(branch, this);
        Q_UNUSED(glog);
    }
    else
        reset();
}

void ChangesetModel::branchLogRead(GetBranchLog *glog)
{
    QString branch = glog->getBranch();
    branchMap[branch].revisions << glog->getRevisions();
    if(branch == currentBranch)
    {
        receiveRevisions(false);
    }
    delete glog;
}

void ChangesetModel::changesetCertReady(ChangesetCerts *cert)
{
    QDateTime date;
    QString author;
    QString changelog;
    QString key;
    QString value;
    int rows = cert->rowCount(QModelIndex());
    for(int i = 0; i < rows; i++)
    {
        key = cert->data(cert->index(i, 1, QModelIndex()), Qt::DisplayRole).toString();
        value = cert->data(cert->index(i, 2, QModelIndex()), Qt::DisplayRole).toString();
        if(key == "author")
        {
            author = value;
            continue;
        }
        if(key == "changelog")
        {
            changelog = value;
            continue;
        }
        if(key == "date")
        {
            date = QDateTime::fromString(value, "yyyy-MM-ddThh:mm:ss");
            continue;
        }
    }
    Changeset *set = new Changeset(cert->getRevision(), date, author, changelog);
    changesetMap[cert->getRevision()] = set;
    int idx = cert->getIndex();
    branchMap[currentBranch].certsRead = idx;
    beginInsertRows(QModelIndex(), idx, idx);
    endInsertRows();
    delete cert;
}

int ChangesetModel::columnCount(const QModelIndex &parent) const
{
    return 4;
}

QVariant ChangesetModel::data(const QModelIndex &index, int role) const
{
    if (!index.isValid())
    {
        return QVariant();
    }
    
    if (role == Qt::FontRole)
    {
        QFont font;
        font.setStyleHint(QFont::Courier);
        font.setFamily("Courier");
        return QVariant(font);
    }
    
    if (role == Qt::DisplayRole)
    {
        BranchInfo branchInfo = branchMap[currentBranch];
        int row = index.row();
        if (branchInfo.revisions.count() > row)
        {
            int col = index.column();
            QString rev = branchInfo.revisions[row];
            if(changesetMap[rev] == NULL) return QVariant();
            switch(col)
            {
            case(0):
                return QVariant(changesetMap[rev]->getDate());
            case(1):
                return QVariant(changesetMap[rev]->getAuthor());
            case(2):
                return QVariant(changesetMap[rev]->getChangelogFlat());
            case(3):
                return QVariant(rev);
            }
        }
    }
    
    return QVariant();
}

Qt::ItemFlags ChangesetModel::flags(const QModelIndex &index) const
{
    if (index.isValid())
    {
        return Qt::ItemIsEnabled | Qt::ItemIsSelectable;
    }
    return 0;
}

QVariant ChangesetModel::headerData(int section, Qt::Orientation orientation, int role) const
{
    if (orientation == Qt::Horizontal && role == Qt::DisplayRole)
    {
        switch(section)
        {
        case(0):
            return QVariant(tr("Date"));
        case(1):
            return QVariant(tr("Author"));
        case(2):
            return QVariant(tr("Changelog"));
        case(3):
            return QVariant(tr("Revision ID"));
        }
    }
    return QVariant();
}

int ChangesetModel::rowCount(const QModelIndex& parent) const
{
    int i = branchMap[currentBranch].certsRead;
    if (!parent.isValid())
        if(branchMap[currentBranch].revisions.count() != 0)
            return i;
    return 0;
}

QModelIndex ChangesetModel::index(int row, int column, const QModelIndex& parent) const
{
    if (!hasIndex(row, column, parent))
    {
        return QModelIndex();
    }
    
    return createIndex(row, column, 0);
}

QModelIndex ChangesetModel::parent(const QModelIndex& index) const
{
    return QModelIndex();
}

