/***************************************************************************
 *   Copyright (C) 2006 by Thomas Keller                                   *
 *   me@thomaskeller.biz                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
 
#include "ContentDiff.h"
#include "Guitone.h"

#include <QFont>
#include <QBrush>

ContentDiff::ContentDiff(QObject *parent) 
    : QAbstractItemModel(parent)
{
    diffParser = 0;
    mtnDelegate = new MonotoneDelegate(this);
}

ContentDiff::~ContentDiff()
{
    if (diffParser) delete diffParser;
	qDeleteAll(lines);
    lines.clear();
    delete mtnDelegate;
}

bool ContentDiff::readDiff(QString fileName)
{
    return readDiff(fileName, QString(), QString());
}

bool ContentDiff::readDiff(QString fileName, QString leftRevision)
{
    return readDiff(fileName, leftRevision, QString());
}

bool ContentDiff::readDiff(
    QString fileName, QString leftRevision, QString rightRevision
)
{
    // reset the view
    reset();
    
    QStringList cmd;
    cmd << "content_diff" << fileName;
    
    QStringList opts;
    
    if (leftRevision.length() == 0)
    {
        leftRevision = MonotoneDelegate::getBaseWorkspaceRevision(this);
    }
    
    opts << "r" << leftRevision;
    
    if (rightRevision.length() > 0)
    {
        opts << "r" << rightRevision;
    }
    
    return mtnDelegate->triggerCommand(cmd, opts);
}

void ContentDiff::parseOutput()
{
    diffParser = new DiffParser(AutomateCommand::data);
    
    // flatten the data for the current view
    FileDiffs fileDiffs = diffParser->getAllDiffs();
    
    QMapIterator<QString, Diff*> i(fileDiffs);
    while (i.hasNext())
    {
        i.next();
        ListLine * fileLine = new ListLine(ListLine::FileLine);
        lines.append(fileLine);
        
        Diff * diff = (Diff*)i.value();
        
        if (diff->is_binary)
        {
            fileLine->secondColumn = tr("%1 (binary)").arg(i.key());
            continue;
        }
        
        fileLine->secondColumn = 
            tr("%1 (%2 hunks)").arg(i.key()).arg(diff->hunks.size());
        
        for (int j=0, k=diff->hunks.size(); j<k; j++)
        {
            DiffHunk * hunk = diff->hunks.at(j);
            for (int m=0, n=hunk->lines.size(); m<n; m++)
            {
                DiffLine * line = hunk->lines.at(m);
                ListLine * hunkLine = new ListLine(ListLine::HunkLine);
                fileLine->addChild(hunkLine);
                
                if (line->state == DiffLine::Unchanged)
                {
                    hunkLine->firstColumn = QString::number(hunk->leftStart + m);
                }
                else if (line->state == DiffLine::Added)
                {
                    hunkLine->firstColumn = QString("+");
                }
                else
                {
                    hunkLine->firstColumn = QString("-");
                }
                hunkLine->secondColumn = line->content;
            }
            
            if (j < k-1)
            {
                ListLine * sep = new ListLine(ListLine::HunkSep);
                fileLine->addChild(sep);
            }
        }
    }
    
    reset();
    
    emit diffRead();
}

int ContentDiff::columnCount(const QModelIndex &parent) const
{
	return 2;
}

QVariant ContentDiff::data(const QModelIndex &index, int role) const
{
    if (!index.isValid()) return QVariant();
    
    int col = index.column();
    ListLine * line = static_cast<ListLine*>(index.internalPointer());
    
    if (role == Qt::DisplayRole)
    {
        switch (col)
        {
            case 0: return QVariant(line->firstColumn);
            case 1: return QVariant(line->secondColumn);
        }
        
        I(false);
    }
    
    if (role == Qt::FontRole && col == 1)
    {
        QFont font;
        
        if (line->type == ListLine::FileLine)
        {
            font.setBold(true);
        }
        
        if (line->type == ListLine::HunkLine)
        {
            font.setStyleHint(QFont::Courier);
            font.setFamily("Courier");
        }
        
        return QVariant(font);
    }
    
    if (role == Qt::TextAlignmentRole && col == 0)
    {
        return QVariant(Qt::AlignRight);
    }
    
    if (role == Qt::BackgroundRole)
    {
        if (line->type != ListLine::HunkLine) return QVariant();
        
        if (line->firstColumn.compare("-") == 0)
        {
            return QVariant(QBrush(Qt::red));
        }
        
        if (line->firstColumn.compare("+") == 0)
        {
            return QVariant(QBrush(Qt::green));
        }
        
        // unchanged lines
        return QVariant();
    }
    
    return QVariant();
}

Qt::ItemFlags ContentDiff::flags(const QModelIndex &index) const
{    
	return Qt::ItemIsEnabled | Qt::ItemIsSelectable;
}

QVariant ContentDiff::headerData(int section, Qt::Orientation orientation, int role) const
{
	if (role != Qt::DisplayRole || orientation != Qt::Horizontal)
        return QVariant();
    if (section == 0) return QVariant(tr("Line"));
    return QVariant(tr("File/Content"));
}

int ContentDiff::rowCount(const QModelIndex& parent) const
{
    if (!parent.isValid()) return lines.size();
    ListLine * line = static_cast<ListLine*>(parent.internalPointer());
    return line->lines.size();
}

QModelIndex ContentDiff::index(int row, int column, const QModelIndex& parent) const
{
    if (!parent.isValid())
    {
        if (row >= lines.size()) return QModelIndex();
        ListLine * line = lines.at(row);
        return createIndex(row, column, line);
    }

    ListLine * parentLine = static_cast<ListLine*>(parent.internalPointer());
    if (row >= parentLine->lines.size()) return QModelIndex();
    ListLine * line = parentLine->lines.at(row);
    return createIndex(row, column, line);
}

QModelIndex ContentDiff::parent(const QModelIndex& index) const
{
    if (!index.isValid()) return QModelIndex();
    ListLine * line = static_cast<ListLine*>(index.internalPointer());
    if (!line->parent) return QModelIndex();
    return createIndex(lines.indexOf(line->parent) , 0, line->parent);
}

