/***************************************************************************
*   Copyright (C) 2006 by Thomas Keller                                   *
*   me@thomaskeller.biz                                                   *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
***************************************************************************/

#include "Inventory.h"
#include "InventoryItem.h"
#include "Guitone.h"
#include "IconProvider.h"

#include <QApplication>

Inventory::Inventory(QObject *parent) : QAbstractItemModel(parent)
{
	// create a dummy item since the view needs at least one item
	// in the model, otherwise the app crashes
	rootItem	 = new InventoryItem();
    regex		 = new QRegExp("^(R|D|[ ])(R|A|[ ])(M|P|U|I|[ ])\\s(\\d+)\\s(\\d+)\\s(.+)$");
    regex->setMinimal(true);
    
    mtnDelegate  = new MonotoneDelegate(this);
    
    connect(
        this, SIGNAL(modelCreated()),
        this, SLOT(loadBranchName())
    );
}

Inventory::~Inventory()
{
	delete rootItem;
	delete regex;
    delete mtnDelegate;
}

bool Inventory::readInventory()
{
    QStringList cmd;
    cmd << "inventory";

	return mtnDelegate->triggerCommand(cmd);
}

void Inventory::parseOutput()
{
	QStringList lines = AutomateCommand::data.split("\n", QString::SkipEmptyParts);
	QMap<int, InventoryItem*> renameMap;
	QMap<int, InventoryItem*>::iterator renameIter;

	InventoryItem *item;
	QList<InventoryItem*> tempItems;
	int status(0);
	int from_id(0);
	int to_id(0);
	QString path("");
	bool isDirectory(false);

	for (QStringList::Iterator it = lines.begin(); it != lines.end(); ++it)
	{
		if (!parseInventoryLine(*it, status, from_id, to_id, path, isDirectory))
		{
			continue;
		}
		
		// this item is given a parent explicitely later on
		item = new InventoryItem(isDirectory);
        item->setPath(path);
        item->setStatus(status);
        
		if (from_id > 0)
		{
			renameMap[-from_id] = item;
		}
		if (to_id > 0)
		{
			renameMap[to_id] = item;
		}

		tempItems.push_back(item);
	}

	int id = 0;

	while (true)
	{
		renameIter = renameMap.find(++id);
		if (renameIter == renameMap.end()) break;

		renameMap[id]->setRenamedFrom(renameMap[-id]);
		renameMap[-id]->setRenamedTo(renameMap[id]);
	}

    // FIXME: we shouldn't really add a workspace root item here, but
    // mtn automate inventory currently doesn't print the root workspace dir
    InventoryItem * branch = new InventoryItem(true, true);
    branch->setParent(rootItem);
    branch->setPath(".");
    branch->setStatus(0); 
    branch->setChildren(buildTreeRecursive(tempItems, NULL));
    
    rootItem->appendChild(branch);
    
	// reset the model to repaint the view completly
	// (all QModelIndexes are discarded through that, e.g. selections!)
	reset();

	// restore the normal cursor
	qApp->restoreOverrideCursor();

    emit modelCreated();
}

void Inventory::loadBranchName()
{
#ifdef Q_WS_WIN32
    W("Querying branch name disabled, will lead to program halt.");
#else
    QList<InventoryItem*> children = rootItem->getChildren();
    I(children.size() > 0);
    children[0]->setLabel(MonotoneDelegate::getBranchName(this));
    emit layoutChanged();
#endif
}

QList<InventoryItem*> Inventory::buildTreeRecursive(QList<InventoryItem*> &items, InventoryItem* parentItem)
{
	QList<InventoryItem*> finalItems;

	QString parentPath = "";
    int parentStatus = 0;

	if (parentItem != NULL)
	{
		parentPath = parentItem->getPath();
        parentStatus = parentItem->getStatus();
	}
    
    // add pseudo item "cd up" for each directory
    InventoryItem * cdUp = new InventoryItem(true);
    cdUp->setParent(parentItem);
    cdUp->setPath(parentPath + QString("/.."));

    items.prepend(cdUp);

    InventoryItem  * currentItem;
	while (items.size() > 0)
	{
		currentItem = items.front();

		//
		// if the item is not directly inside the current path stop immediately
		//
		QString tmpPath = parentPath;
		tmpPath.append("/");
		tmpPath.append(currentItem->getFilename());

		// path may not be empty for this condition
		if (!parentPath.isEmpty() && currentItem->getPath().compare(tmpPath) != 0)
		{
			break;
		}

		// remove the item if we can append it now
		items.pop_front();

		//
		// it seems to be a valid item
		//

		// if the item is directory a directory, make sure we catch decendant items... recursion!
		if (currentItem->isDirectory() && !currentItem->isCdUp())
		{
			currentItem->setChildren(buildTreeRecursive(items, currentItem));
        }

		// append item to final list
		finalItems.push_back(currentItem);

	}
	return finalItems;
}

QModelIndex Inventory::index(int row, int column, const QModelIndex &parent)
const
{
	InventoryItem * parentItem;

	if (!parent.isValid())
	{
		parentItem = rootItem;
	}
	else
	{
		parentItem = static_cast<InventoryItem*>(parent.internalPointer());
	}

	InventoryItem * childItem = parentItem->child(row);

	if (childItem)
	{
		return createIndex(row, column, childItem);
	}

	return QModelIndex();
}

int Inventory::columnCount(const QModelIndex &parent) const
{
	if (parent.isValid())
	{
		return static_cast<InventoryItem*>(parent.internalPointer())->columnCount();
	}

	return rootItem->columnCount();
}

QVariant Inventory::data(const QModelIndex &index, int role) const
{
	if (!index.isValid())
	{
		return QVariant();
	}

	InventoryItem * item = static_cast<InventoryItem*>(index.internalPointer());

	if ((role == Qt::DecorationRole) && (index.column() == 0))
	{
        IconProvider * provider = IconProvider::singleton();
		return provider->getIcon(item);
	}
	else
	{
		return item->data(index.column(), role);
	}
}

bool Inventory::setData(const QModelIndex & idx, const QVariant & value, int role)
{
    return false;
}

bool Inventory::setItemData(const QModelIndex & index, const QMap<int, QVariant> & roles)
{
    return false;
}

Qt::ItemFlags Inventory::flags(const QModelIndex &index) const
{
	if (!index.isValid()) return 0;
    
    QFlags<Qt::ItemFlag> flags = Qt::ItemIsSelectable | Qt::ItemIsEnabled;

    InventoryItem * item = static_cast<InventoryItem*>(index.internalPointer());
    
    if (item->isCdUp() || item->isRootDirectory()) return flags;
    
    /*
    // Disabled until we have figured out how to implement renaming properly
    flags |= Qt::ItemIsEditable | Qt::ItemIsDragEnabled;
    
    if (item->isDirectory())
	{
		flags |= Qt::ItemIsDropEnabled;
    }
	*/
    
    return flags;
}

// the root item needs to store the header for each field
QVariant Inventory::headerData(int section, Qt::Orientation orientation, int role) const
{
	if (orientation == Qt::Horizontal && role == Qt::DisplayRole)
	{
		return rootItem->data(section, role);
	}

	return QVariant();
}

QModelIndex Inventory::parent(const QModelIndex& index) const
{
	if (!index.isValid())
	{
		return QModelIndex();
	}

	InventoryItem * childItem = static_cast<InventoryItem*>(index.internalPointer());
	InventoryItem * parentItem = childItem->parent();

	if (parentItem == rootItem)
	{
		return QModelIndex();
	}

	return createIndex(parentItem->row(), 0, parentItem);
}

int Inventory::rowCount(const QModelIndex& parent) const
{
	InventoryItem * parentItem = rootItem;

	if (parent.isValid())
	{
		parentItem = static_cast<InventoryItem*>(parent.internalPointer());
	}

	return parentItem->childCount();
}

bool Inventory::parseInventoryLine(
	const QString &inputString,
	int &status,
	int &from_id, 
	int &to_id, 
	QString &path,
	bool &isDirectory)
{
	if (regex->indexIn(inputString) == -1)
	{
		W(QString("Couldn't parse inventory line %1").arg(inputString));
		return false;
	}

	QStringList list = regex->capturedTexts();
	status = 0;

	// the first match
	if (list[1].compare("R") == 0)
	{
		status |= InventoryItem::RenamedFrom;
	} else
	if (list[1].compare("D") == 0)
	{
		status |= InventoryItem::Dropped;
	}
	else
	if (list[1].compare(" ") != 0)
	{
		W(QString("Unknown status first tripel %1").arg(list[1]));
	}


	// the second match
	if (list[2].compare("R") == 0)
	{
		status |= InventoryItem::RenamedTo;
	} else
	if (list[2].compare("A") == 0)
	{
		status |= InventoryItem::Added;
	}
	else
	if (list[2].compare(" ") != 0)
	{
		W(QString("Unknown status second tripel %1").arg(list[2]));
	}

	// the third match
	if (list[3].compare("M") == 0)
	{
		status |= InventoryItem::Missing;
	} else
	if (list[3].compare("P") == 0)
	{
		status |= InventoryItem::Patched;
	} else
	if (list[3].compare("U") == 0)
	{
		status |= InventoryItem::Unknown;
	} else
	if (list[3].compare("I") == 0)
	{
		status |= InventoryItem::Ignored;
	} 
    else
	if (list[3].compare(" ") == 0)
	{
		status |= InventoryItem::Unchanged;
	} 
    else
	{
        W(QString("Unknown status third tripel %1").arg(list[3]));
	}
    
    I(InventoryItem::ValidStates.contains(status));
    
	// now determine if the file has been renamed
	from_id = list[4].toInt();
	to_id   = list[5].toInt();

	// remove trailing slash
	path = list[6].trimmed();
    
    isDirectory = false;
	if (path.endsWith('/'))
	{
		isDirectory = true;
		path = path.left(path.length() - 1);
	}
	
	// parsing was successful
	return true;
}

bool Inventory::handleError(int errCode)
{
    if (errCode != 2) return false;
    
    // FIXME: normally there should only be an error if the workspace format
    // itself is not readable (i.e. wrong monotone version for the given
    // workspace version), but who knows what could bomb out here as well. in the
    // meantime we ensure the user gets a proper error message. in the
    // future we could do the task of upgrading a workspace / database
    // format right here in guitone
    // Starting with mtn 0.33 invalid databases are catched as early as 
    // automate stdio is run, only for 0.32 and below the database is checked
    // on the very first command.
    
    emit invalidWorkspaceFormat(Monotone::stripMtnPrefix(AutomateCommand::data));
    return true;
}

