/***************************************************************************
 *   Copyright (C) 2006 by Thomas Keller                                   *
 *   me@thomaskeller.biz                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "InventoryItem.h"
#include "vocab.h"

#include <QStringList>
#include <QFont>

const int InventoryItem::RenamedFrom = 1;
const int InventoryItem::RenamedTo = 2;
const int InventoryItem::Added = 4;
const int InventoryItem::Dropped = 8;
const int InventoryItem::Missing = 16;
const int InventoryItem::Patched = 32;
const int InventoryItem::Unchanged = 64;
const int InventoryItem::Unknown = 128;
const int InventoryItem::Ignored = 256;

//
// initialize the array with all 26 valid states (out of 45 possible)
//
const QList<int> InventoryItem::ValidStates = QList<int>() 
    << InventoryItem::Unchanged                                                            // '   '
    << InventoryItem::Patched                                                              // '  P'
    << InventoryItem::Unknown                                                              // '  U'
    << InventoryItem::Ignored                                                              // '  I'
    << InventoryItem::Missing                                                              // '  M'
    << ( InventoryItem::Added | InventoryItem::Patched )                                   // ' AP'
    << ( InventoryItem::Added | InventoryItem::Missing )                                   // ' AM'
    << ( InventoryItem::RenamedTo | InventoryItem::Unchanged )                             // ' R '
    << ( InventoryItem::RenamedTo | InventoryItem::Patched )                               // ' RP'
    << ( InventoryItem::RenamedTo | InventoryItem::Missing )                               // ' RM'
    << ( InventoryItem::Dropped | InventoryItem::Unchanged )                               // 'D  '
    << ( InventoryItem::Dropped | InventoryItem::Unknown )                                 // 'D U'
    << ( InventoryItem::Dropped | InventoryItem::Ignored )                                 // 'D I'
    << ( InventoryItem::Dropped | InventoryItem::Added | InventoryItem::Patched )          // 'DAP'
    << ( InventoryItem::Dropped | InventoryItem::Added | InventoryItem::Missing )          // 'DAM'
    << ( InventoryItem::Dropped | InventoryItem::RenamedTo | InventoryItem::Unchanged )    // 'DR '
    << ( InventoryItem::Dropped | InventoryItem::RenamedTo | InventoryItem::Patched )      // 'DRP'
    << ( InventoryItem::Dropped | InventoryItem::RenamedTo | InventoryItem::Missing )      // 'DRM'
    << ( InventoryItem::RenamedFrom | InventoryItem::Unchanged )                           // 'R  '
    << ( InventoryItem::RenamedFrom | InventoryItem::Unknown )                             // 'R U'
    << ( InventoryItem::RenamedFrom | InventoryItem::Ignored )                             // 'R I'
    << ( InventoryItem::RenamedFrom | InventoryItem::Added | InventoryItem::Patched )      // 'RAP'
    << ( InventoryItem::RenamedFrom | InventoryItem::Added | InventoryItem::Missing )      // 'RAM'
    << ( InventoryItem::RenamedFrom | InventoryItem::RenamedTo | InventoryItem::Unchanged )// 'RR '
    << ( InventoryItem::RenamedFrom | InventoryItem::RenamedTo | InventoryItem::Patched )  // 'RRP'
    << ( InventoryItem::RenamedFrom | InventoryItem::RenamedTo | InventoryItem::Missing ); // 'RRM'

InventoryItem::InventoryItem(bool isDir /* =false */, bool isRoot /* =false */)
{
    parentItem = this;
    path = "";
    label = "";
    status = 0;
    dirFlag = isDir;
	rootFlag = isRoot;
}

InventoryItem::InventoryItem(const InventoryItem * other)
{
    dirFlag = other->isDirectory();
    rootFlag = other->isRootDirectory();
    renamed_from = other->getRenamedFrom();
    renamed_to = other->getRenamedTo();
    parentItem = other->parent();
    children = other->getChildren();
    status = other->getStatus();
    path = other->getPath();
    label = other->getLabel();
}

InventoryItem::~InventoryItem()
{
    qDeleteAll(children);
    children.clear();
}

QString InventoryItem::getItemName() const
{
    if (label.size() > 0)
    {
        return label;
    }
    
    return getFilename();
}

void InventoryItem::deleteAllChildren(void)
{
	parentItem = this;
    path = "";
    status = 0;
    dirFlag = false;
    rootFlag = false;
	qDeleteAll(children);
	children.clear();
}

void InventoryItem::appendChild(InventoryItem* child)
{
	child->setParent(this);
    children.append(child);
}

void InventoryItem::setChildren(QList<InventoryItem*> items)
{
	// set the current element as parent for each children
	QList<InventoryItem*>::iterator iter;

	InventoryItem *item;
	for (iter = items.begin(); iter != items.end(); ++iter)
	{
		item = (*iter);
		item->setParent(this);
	}
	
	children.clear();
    children = items;
}

int InventoryItem::row() const
{
    if (parentItem)
        return parentItem->children.indexOf(const_cast<InventoryItem*>(this));

    return 0;
}

QVariant InventoryItem::data(int column, int role) const
{
	if (role == Qt::DisplayRole)
	{
		// return column headers for root item
		if (parentItem == this)
		{
			switch (column)
			{
                case 0: return QVariant(QString(tr("File")));
                case 1: return QVariant(QString(tr("Status")));
                case 2: return QVariant(QString(tr("Additional info")));
                default: return QVariant();
			}
		}

        switch (column)
		{
            case 0: return QVariant(getItemName());
            case 1: return QVariant(getStatusString());
            case 2: return QVariant(getRenameInfo());
            default: return QVariant();
		}
	}
	else if (role == Qt::FontRole && column == 0) 
	{
        QFont font;
        font.setBold(false);
        if (hasChangedRecursive())
        {
            font.setBold(true);
        }
        return QVariant(font);
	}
	else
	{
		return QVariant();
	}
}

QString InventoryItem::getFilename() const
{
    int pos = path.lastIndexOf('/');
    return pos == -1 ? path : path.right(path.length() - pos - 1);
}

QString InventoryItem::getRelativePath(const QString & part) const
{
    int partLen = part.length();
    I(path.left(partLen).compare(part) == 0);
    return path.right(path.length() - partLen);
}

QString InventoryItem::getBaseDirectory() const
{
    int pos = path.lastIndexOf('/');
    return pos == -1 ? QString("") : path.left(pos + 1);
}

bool InventoryItem::hasStatus(int statusBits) const
{
	return (status & statusBits) == statusBits;
}

bool InventoryItem::hasNotStatus(int statusBits) const
{
	return (status & statusBits) == 0;
}

int InventoryItem::getStatusRecursive() const
{
    int overallStatus = status;
    
    if (!isDirectory())
    {
        return overallStatus;
    }
    
    for (int i=0,s=children.size(); i<s; i++)
    {
        InventoryItem * item = static_cast<InventoryItem*>(children.at(i));
        overallStatus |= item->getStatusRecursive();
    }
    
    
    return overallStatus;
}

bool InventoryItem::hasChangedRecursive() const
{
    int state = getStatusRecursive();
    return 
        (state & Added) == Added ||
        (state & Dropped) == Dropped ||
        (state & RenamedFrom) == RenamedFrom ||
        (state & RenamedTo) == RenamedTo ||
        (state & Patched) == Patched;
}

QString InventoryItem::getStatusString() const
{
    // do not return the status for  
    if (isCdUp()) return "";
    
    QStringList list;
    
    // at first the possible old states
    if (this->hasStatus(InventoryItem::RenamedFrom))
    {
        list.append(tr("Rename Source"));
    }
    
    if (this->hasStatus(InventoryItem::Dropped))
    {
        list.append(tr("Dropped"));
    }
    
    // now the possible new states
    if (this->hasStatus(InventoryItem::RenamedTo))
    {
        list.append(tr("Rename Target"));
    }
    if (this->hasStatus(InventoryItem::Added))
    {
        list.append(tr("Added"));
    }
    
    // and finally the file states
    if (this->hasStatus(InventoryItem::Missing))
    {
        list.append(tr("Missing"));
    }
    if (this->hasStatus(InventoryItem::Patched))
    {
        list.append(tr("Modified"));
    }
    if (this->hasStatus(InventoryItem::Unchanged))
    {
        list.append(tr("Unchanged"));
    }
    if (this->hasStatus(InventoryItem::Unknown))
    {
        list.append(tr("Unknown"));
    }
    if (this->hasStatus(InventoryItem::Ignored))
    {
        list.append(tr("Ignored"));
    }
    
    return list.join(", ");
}

QString InventoryItem::getRenameInfo() const
{
    QStringList strings;
    if (hasStatus(RenamedFrom))
    {
        InventoryItem * item = getRenamedTo();
        I(item);
        strings << tr("new name: %1").arg(item->getPath());
    }
    
    if (hasStatus(RenamedTo))
    {
        InventoryItem * item = getRenamedFrom();
        I(item);
        strings << tr("old name: %1").arg(item->getPath());
    }
    return strings.join(", ");
}

