/***************************************************************************
*   Copyright (C) 2006 by Ingo Maindorfer								   *
*   ingo@liquidcooling.de												   *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
***************************************************************************/

#include "InventoryProxyModel.h"
#include "InventoryItem.h"

InventoryProxyModel::InventoryProxyModel(QObject *parent, bool folderTree_) 
: QSortFilterProxyModel(parent), folderTree(folderTree_),
sortColumn(0), hideIgnored(false), sortOrder(Qt::AscendingOrder)
{
}

InventoryProxyModel::~InventoryProxyModel(void)
{
}

bool InventoryProxyModel::filterAcceptsRow(int sourceRow, const QModelIndex &sourceParent) const
{
	QModelIndex index = sourceModel()->index(sourceRow, 0, sourceParent);
	InventoryItem *item = static_cast<InventoryItem*>(index.internalPointer());

    // display the pseudo CdUp item only in the file list regardless of the
    // state of the item
    if (item->isCdUp())
    {
        // only display this item if this is not the folder tree view
        // and the parent of the item's parent is invalid (i.e. there is no
        // way to cdUp anymore)
        return !folderTree && sourceParent.parent().isValid();
    }
    
    // hide ignored files (not recursively)
    if (item->hasStatus(InventoryItem::Ignored) && hideIgnored)
    {
        return false;
    }
    
    bool acceptRow = true;
    
    // check the view options
    int state    = item->getStatusRecursive();
    bool patched = state & InventoryItem::Patched;
    bool added   = state & InventoryItem::Added;
    bool dropped = state & InventoryItem::Dropped;
    bool renamed = state & InventoryItem::RenamedFrom || state & InventoryItem::RenamedTo;
    bool unknown = state & InventoryItem::Unknown;
    bool missing = state & InventoryItem::Missing;
    bool ignored = state & InventoryItem::Ignored;
    bool changed = patched || added || dropped || renamed; 
    
    acceptRow &= viewOption != Changed || changed;
    acceptRow &= viewOption != Patched || patched;
    acceptRow &= viewOption != Added   || added;
    acceptRow &= viewOption != Dropped || dropped;
    acceptRow &= viewOption != Renamed || renamed;
    acceptRow &= viewOption != Unknown || unknown;
    acceptRow &= viewOption != Missing || missing;
    acceptRow &= viewOption != Ignored || ignored;
    
    // check if we should only display folders 
    acceptRow &= !folderTree || item->isDirectory();
    return acceptRow;
}

bool InventoryProxyModel::lessThan(const QModelIndex &left, const QModelIndex &right) const
{
	// TODO: Needs some tweaking for correct sorting when clicked on different headers
	// f.e. status strings are sorted differently based on their translation
	InventoryItem *itemLeft = static_cast<InventoryItem*>(left.internalPointer());
	InventoryItem *itemRight = static_cast<InventoryItem*>(right.internalPointer());

	// We shall sort by the name of the item
	if (sortColumn == 0) 
	{
        // place cd up items always on the first position
        if (itemLeft->isCdUp() || itemRight->isCdUp()) return false;
        
		if (itemLeft->isDirectory() && itemRight->isDirectory())
		{
			return itemLeft->getFilename() < itemRight->getFilename();
		}

		if (!itemLeft->isDirectory() && !itemRight->isDirectory())
		{
			return itemLeft->getFilename() < itemRight->getFilename();
		}

		if (itemLeft->isDirectory() && !itemRight->isDirectory())
		{
			return true;
		}

		// last case: !itemLeft->isDirectory() && itemRight->isDirectory()
		return false;
	}
	// we shall sort only by status
	else if (sortColumn == 1)
	{
		return itemLeft->getStatus() < itemRight->getStatus();
	}
	// anything else
	else
	{
		return false;
	}
}

void InventoryProxyModel::sort(int column, Qt::SortOrder order)
{
    sortOrder = order;
	sortColumn = column;
    QSortFilterProxyModel::sort(column, order);
}

void InventoryProxyModel::setHideIgnoredFiles(bool hide)
{
    if (hide == hideIgnored) return;
    hideIgnored = hide;
    clear();
}

void InventoryProxyModel::setViewOption(ViewOption opt)
{
    if (viewOption == opt) return;
    viewOption = opt;
    clear();
}

