/***************************************************************************
 *   Copyright (C) 2006 by Thomas Keller                                   *
 *   me@thomaskeller.biz                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
 
#include "Keys.h"
#include "Monotone.h"
#include "BasicIOParser.h"

const int Key::Database = 1;
const int Key::Keystore = 2;

Keys::Keys(QObject *parent) : QAbstractItemModel(parent)
{
    mtnDelegate = new MonotoneDelegate(this);
}

Keys::~Keys()
{
    delete mtnDelegate;
}

bool Keys::readKeys()
{
    keys.clear();
    reset();
    
    QStringList cmd;
    cmd << "keys";
    
    return mtnDelegate->triggerCommand(cmd);
}

void Keys::parseOutput()
{
    BasicIOParser parser(AutomateCommand::data);
    I(parser.parse());
    StanzaList list = parser.getStanzas();
    
    for (int i=0, size = list.size(); i < size; ++i)
    {
        Stanza stanza = list.at(i);
        
        Key key;
        bool isItem = false;
        
        for (int j=0, size2 = stanza.size(); j < size2; j++)
        {
            StanzaEntry entry = stanza.at(j);
            
            // we're now only interested in stanzas starting with a "name" entry
            if (j == 0 && entry.sym != "name") break;
            
            isItem = true;
            
            if (entry.sym == "name")
            {
                I(entry.vals.size() == 1);
                key.name = entry.vals.at(0);
                continue;
            }
            
            if (entry.sym == "public_hash")
            {
                I(entry.vals.size() == 1);
                key.public_hash = entry.vals.at(0);
                continue;
            }
            
            if (entry.sym == "private_hash")
            {
                I(entry.vals.size() == 1);
                key.private_hash = entry.vals.at(0);
                continue;
            }
            
            if (entry.sym == "public_location")
            {
                for (int k=0, size3 = entry.vals.size(); k < size3; k++)
                {
                    if (entry.vals.at(k) == "database")
                    {
                        key.public_locations |= Key::Database;
                        continue;
                    }
                    
                    if (entry.vals.at(k) == "keystore")
                    {
                        key.public_locations |= Key::Keystore;
                        continue;
                    }
                    
                    W(QString("Unknown key location '%1'.").arg(entry.vals.at(k)));
                }
                continue;
            }
            
            if (entry.sym == "private_location")
            {
                for (int k=0, size3 = entry.vals.size(); k < size3; k++)
                {
                    if (entry.vals.at(k) == "database")
                    {
                        key.private_locations |= Key::Database;
                        continue;
                    }
                    
                    if (entry.vals.at(k) == "keystore")
                    {
                        key.private_locations |= Key::Keystore;
                        continue;
                    }
                    
                    W(QString("Unknown key location '%1'.").arg(entry.vals.at(k)));
                }
                continue;
            }
            
            W(QString("Unknown symbol %1.").arg(entry.sym));
        }
        
        // check if we really processed an item entry
        if (!isItem) continue;
        keys.append(key);
    }    
    
    reset();
    emit keysRead();
}

int Keys::columnCount(const QModelIndex &parent) const
{
	return 5;
}

QVariant Keys::data(const QModelIndex &index, int role) const
{
	if (!index.isValid())
	{
		return QVariant();
	}
    
    if (role != Qt::DisplayRole)
	{
        return QVariant();
    }
	
    int row = index.row();
    if (row >= keys.size()) return QVariant();
    
    Key key(keys.at(row));
            
    switch (index.column())
    {
        case 0: return QVariant(key.name);
        case 1: return QVariant(key.public_hash);
        case 2: return QVariant(key.private_hash);
        case 3: return QVariant(getLocationString(key.public_locations));
        case 4: return QVariant(getLocationString(key.private_locations));
    }
    
    return QVariant();
}

QString Keys::getLocationString(int loc) const
{
    QStringList str;
    if ((loc & Key::Database) == Key::Database)
        str << tr("Database");
    if ((loc & Key::Keystore) == Key::Keystore)
        str << tr("Keystore");
    return str.join(", ");
}

Qt::ItemFlags Keys::flags(const QModelIndex &index) const
{    
	return Qt::ItemIsEnabled | Qt::ItemIsSelectable;
}

QVariant Keys::headerData(int section, Qt::Orientation orientation, int role) const
{
	if (orientation == Qt::Horizontal && role == Qt::DisplayRole)
	{
        switch (section)
        {
            case 0: return QVariant(tr("Name"));
            case 1: return QVariant(tr("Public Hash"));
            case 2: return QVariant(tr("Private Hash"));
            case 3: return QVariant(tr("Public Locations"));
            case 4: return QVariant(tr("Private Locations"));
        }
    }
    return QVariant();
}

int Keys::rowCount(const QModelIndex& parent) const
{
	return keys.size();
}

QModelIndex Keys::index(int row, int column, const QModelIndex& parent) const
{
    if (!hasIndex(row, column, parent))
    {
        return QModelIndex();
    }
    
    if (row >= keys.size()) return QModelIndex();
    Key key(keys.at(row));
    return createIndex(row, column, &key);
}

QModelIndex Keys::parent(const QModelIndex& index) const
{
    return QModelIndex();
}

