/***************************************************************************
 *   Copyright (C) 2006 by Thomas Keller                                   *
 *   me@thomaskeller.biz                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
 
// 
// There are two ways to query data from the running monotone process:
// asynchronous and synchronous. Asynchronous queries are mainly useful for
// commands which can take a variable time to finish (e.g. inventory), while
// synchronous queries are good for constant time queries (get_base_revision_id
// and friends), whose data are actually prerequisites for more complex commands.
// 
// In any of both cases a monotone process needs to be started beforehand by
// running setup() which takes a workspace directory as argument.
// 
// The basic code flow is then the following:
// 
//      Asynchronous                             Synchronous
//      ------------                             -----------
// 
//     triggerCommand()                        executeCommand()
//           |                                        |
//           V                                        V
//       writeStdin()                            writeStdin()
//          ...                                       |
//          ...                                       V
//     wait for signalling, readAndParseStdout() and put it into the
//     output map (mapped by command number) - if we receive an "end
//     command" token, signal that the next command has been finished
//          ...                                       |
//          ...                                       V
//      signal the caller                  catch the signal locally
//     to read out the data                and immediately return to
//                                                the caller
// 
// Since there is only one instance of the monotone process running at a time,
// pending commands need to be queued somehow in order to avoid confusion
// (data are sent to the wrong caller, etc.) - this is done by assigning each
// caller a command number by which it can evaluate if the commandFinished()
// signal it just catched was for the command which it triggered or not
// (using isCommandCompleted(commandNumber)
// 
// If a command is about to be executed, its data are written to STDIN of the
// running process, whilst encoded into stdio format.
// As soon as data arrive, they're read and parsed in readAndParseStdout().
// It might be possible that not all data are written at once to STDOUT of the
// process, so we're careful and just wait for more output if we can't parse
// it at the right time.
// 
// After parsing the output is made available to the calling process by 
// get[Raw|Decoded]Data() - these functions also remove the cached data for the
// command so they should always be called, regardless if output is expected
// or even further processed.
// The getReturnCode() method returns the return code for the specified command
// and resets the command as well (isCommandCompleted() then returns false, if
// this is needed at a later time, one could think about keeping this info, 
// though it should not make a difference now)
//
 
#include "Monotone.h"
#include "SignalWaiter.h"
#include "StdioParser.h"
#include "Guitone.h"

#include <QTextStream>
#include <QApplication>
#include <QTimer>
#include <QDir>
#include <QMutexLocker>

//
// this should be understood as either/or:
// - if the program version is sufficient, everything is fine
// - if not, we additionally check the interface version in case the user
//   compiled his own version with all needed commands
//
const QString Monotone::RequiredProgramVersion = "0.34";
const QString Monotone::RequiredInterfaceVersion = "4.1";

const int Monotone::StdioBufferSize = 50 * 1024 * 1024;
const int Monotone::WaitForMonotoneStart = 15000; // milliseconds

Monotone::Monotone(QObject * parent) : QObject(parent), process(0)
{
    // the count of the commands fired at the current process instance 
	commandCounter = -1;
    // true if the process is destroyed in the destructor
    isCleanExit = false;
    // set to true to abort any running requests
    // FIXME: we certainly need a better way to abort requested data than
    // setting some flag which gets eventually evaluated or not
    doAbortRequests = false;
    // exec timeout waiter
    timeout = false;
    // inialize the process var
    process = 0;
    // path to the monotone binary
    mtnBinaryPath = "";
    // the current working directory (if in workspace mode)
    workDir = "";
    // the current database file (if in database mode)
    databaseFile = "";
}

Monotone::~Monotone()
{
    shutdownCurrentProcess();
}

bool Monotone::normalizeWorkspacePath(QString & workspace) const
{
	QDir tempDir(workspace);
	if (!tempDir.exists())
	{
		W(QString("Cannot find directory %1").arg(workspace));
		return false;
	}

	do
	{
		if (tempDir.cd("_MTN"))
		{
            tempDir.cdUp();
			workspace = tempDir.absolutePath();
			return true;
		}
	}
	while (!tempDir.isRoot() && tempDir.cdUp());

	W(QString("Cannot find _MTN directory in or above %1").arg(workspace));
	return false;
}

bool Monotone::setMtnBinaryPath(const QString & path)
{
    if (checkProgramVersion(path) || checkInterfaceVersion(path))
    {
        mtnBinaryPath = path;
        return true;
    }
    return false;
}

QString Monotone::getNormalizedWorkspacePath() const
{
	return workDir;
}

QString Monotone::getDatabaseFilePath() const
{
    return databaseFile;
}

void Monotone::shutdownCurrentProcess()
{
    // if a previous process is running, exit it cleanly
    if (process)
    {
        isCleanExit = true;
        doAbortRequests = true;
        completedCommands.clear();
        commandCounter = -1;
        
        // close the pipes
        process->close();
        
        // terminate any running process
        process->terminate();
    
        // block until the process has really been finished
        process->waitForFinished();
		
        disconnect(
            process, SIGNAL(readyReadStandardOutput()),
            this, SLOT(readAndParseStdout())
        );

        disconnect(
            process, SIGNAL(readyRead()),
            this, SLOT(readAndParseStdout())
        );
        
        disconnect(
            process, SIGNAL(error(QProcess::ProcessError)),
            this, SLOT(processError(QProcess::ProcessError))
        );

        disconnect(
            process, SIGNAL(finished(int, QProcess::ExitStatus)),
            this, SLOT(processFinished(int, QProcess::ExitStatus))
        );

        delete process;
    }
}

bool Monotone::setupNewProcess()
{
    I(mtnBinaryPath.size() > 0);

    shutdownCurrentProcess();
    
    isCleanExit = false;
    doAbortRequests = false;
    timeout = false;
    
	process = new QProcess(this);

    // monitor if the process is exited unexpectedly
    connect(
        process, SIGNAL(finished(int, QProcess::ExitStatus)),
        this, SLOT(processFinished(int, QProcess::ExitStatus))
    );

    // monitor process errors
    connect(
        process, SIGNAL(error(QProcess::ProcessError)),
        this, SLOT(processError(QProcess::ProcessError))
    );
    
    // parse output from the process
    connect(
        process, SIGNAL(readyRead()),
        this, SLOT(readAndParseStdout())
    );
    
    connect(
        process, SIGNAL(readyReadStandardOutput()),
        this, SLOT(readAndParseStdout())
    );
    
    if (mode == Workspace)
    {
        process->setWorkingDirectory(workDir);
    }
    
    QStringList args;
    args << "automate";
    args << "stdio";
    args << QString("--automate-stdio-size=%1").arg(StdioBufferSize);

    if (mode == Database)
    {
        args << "--db" << databaseFile;
    }
    
    D(QString("starting %1 %2").arg(mtnBinaryPath).arg(args.join(" ")));

    process->start(mtnBinaryPath, args);
    
    if (!process->waitForStarted(WaitForMonotoneStart))
    {
        C("Timed out waiting for monotone start");
        return false;
    }
    
    timer = startTimer(1000);
    
    while (!timeout && !doAbortRequests)
        QCoreApplication::processEvents(QEventLoop::WaitForMoreEvents);
    
    if (!timeout) killTimer(timer);
    
    return !doAbortRequests;
}

void Monotone::timerEvent(QTimerEvent * event)
{
    killTimer(timer);
    timeout = true;
}

bool Monotone::loadWorkspace(QString workspace)
{
    // try to find the _MTN directory in the given path and chdir to the root
    if (!normalizeWorkspacePath(workspace)) return false;
    workDir = workspace;
    mode = Workspace;
    return setupNewProcess();
}

bool Monotone::loadDatabase(QString db)
{
    // FIXME: shall we check here for anything or do we just let monotone
    // escalate later on if something goes wrong, e.g. file is not writable...?
    databaseFile = db;
    mode = Database;
    return setupNewProcess();
}

void Monotone::processError(QProcess::ProcessError error)
{
    if (isCleanExit) return;
    
    // FIXME: a process error can be anything process related. We need to handle
    // this differently and give the user the proper information he needs to
    // resolve the problem.
    // The problem why we don't do that now is that we have no knowledge
    // if the error which occured here will also trigger the finished() signal,
    // which then is handled properly and also contains some process output
    // to throw at the user.
    //
    // There are six different process errors which can come up:
    //
    // QProcess::FailedToStart: 
    //      unlikely, since we check the executable version
    //      before we actually open the stdio process - this could theoretically
    //      popup if the binary is changed / removed between the program start
    //      and the actual start of a stdio process - usually the last workspace
    //      or database is loaded right after start
    // QProcess::Crashed:
    //      we need to determine if this also triggers finished()... if we'd
    //      only have mtn automate crash... and then this should probably be
    //      handled
    // QProcess::Timeout:
    //      occurs on a timeout for any last waitFor...() function - we use
    //      waitForStarted to ensure the process runs before we throw the first
    //      data onto it - maybe this should be handled
    // QProcess::WriteError / QProcess::ReadError:
    //      occur either if a process is not running (crashed? failedToStart?)
    //      or closed its input channel - for any case I believe finished()
    //      is called as well, but this is speculation.
    // QProcess::UnknownError:
    //      yeah, unknown errors, I love these - absolutely no idea what
    //      to do on this or even when it can come up
    C(QString("received process error %1").arg(error));
}

void Monotone::processFinished(int code, QProcess::ExitStatus status)
{
    // this was a normal exit
    if (code == 0 || isCleanExit) return;
    doAbortRequests = true;
}

QString Monotone::getStderr()
{
    if (!process) return QString(tr("[process not created]"));
    QString output(process->readAllStandardError());
    return stripMtnPrefix(output);
}

bool Monotone::executeCommand(const QStringList & command, int & commandNumber)
{
    return executeCommand(command, QStringList(), commandNumber);
}

bool Monotone::executeCommand(const ByteArrayList & command, int & commandNumber)
{
    return executeCommand(command, ByteArrayList(), commandNumber);
}

bool Monotone::executeCommand(const QStringList & command, const QStringList & options, int & commandNumber)
{
    return executeCommand(
        stringToByteArrayList(command), 
        stringToByteArrayList(options),
        commandNumber
    );
}

bool Monotone::executeCommand(const ByteArrayList & command, const ByteArrayList & options, int & commandNumber)
{
    if (doAbortRequests || process->state() != QProcess::Running)
    {
        D("monotone is not running");
	    return false;
    }
    
    commandNumber = writeStdin(command, options);
    SignalWaiter waiter(this, SIGNAL(commandFinished()));
    
    do
    {
        waiter.wait(100);
    }
    while (!doAbortRequests && !isCommandFinished(commandNumber));

    return !doAbortRequests;
}

bool Monotone::triggerCommand(const QStringList & command, int & commandNumber)
{
    return triggerCommand(command, QStringList(), commandNumber);
}

bool Monotone::triggerCommand(const ByteArrayList & command, int & commandNumber)
{
    return triggerCommand(command, ByteArrayList(), commandNumber);
}

bool Monotone::triggerCommand(const QStringList & command, const QStringList & options, int & commandNumber)
{
    return triggerCommand(
        stringToByteArrayList(command), 
        stringToByteArrayList(options),
        commandNumber
    );
}

bool Monotone::triggerCommand(const ByteArrayList & command, const ByteArrayList & options, int & commandNumber)
{
    if (doAbortRequests || process->state() != QProcess::Running)
    {
        D("monotone is not running");
	    return false;
    }
        
    commandNumber = writeStdin(command, options);
    return true;
}

int Monotone::writeStdin(const ByteArrayList & command, const ByteArrayList & options)
{
    QByteArray commandLine;
    QTextStream streamCmdLine(&commandLine);
    
    if (options.size() > 0)
    {
        // currently mtn can only understand key => value option pairs
        I(options.size() % 2 == 0);
        
        streamCmdLine << "o";
        for (int i=0, c=options.size(); i<c; i++)
        {
            QString fragment = options.at(i);
            if (fragment.size() == 0)
            {
                W(QString("skipping empty option %1").arg(i+1));
                continue;
            }
            
            streamCmdLine << fragment.size();
            streamCmdLine << ":";
            streamCmdLine << fragment;
        }
        // separate options from the command by a single whitespace
        streamCmdLine << "e ";
    }
	
    I(command.size() > 0);
    
    streamCmdLine << "l";
    for (int i=0, c=command.size(); i<c; i++)
	{
        QString fragment = command.at(i);
        if (fragment.size() == 0)
        {
            W(QString("skipping empty parameter %1").arg(i+1));
            continue;
        }
            
        streamCmdLine << fragment.size();
        streamCmdLine << ":";
        streamCmdLine << fragment;
    }
	streamCmdLine << "e";
    streamCmdLine.flush();
    
    ++commandCounter;
    
    D(QString("writing command %1 (%2)")
        .arg(commandCounter)
        .arg(commandLine.mid(0,200).data())
    );
    
    QTextStream streamProcess(process);
    streamProcess << commandLine;
    streamProcess.flush();
    
    return commandCounter;
}

void Monotone::readAndParseStdout()
{
    D(QString("read signalled, available bytes: %1").arg(process->bytesAvailable()));
    
    // don't do anything if there is nothing to read
    if (process->bytesAvailable() == 0) return;
    
    // append any new output and try to parse it
    input.append(process->readAllStandardOutput());
    
    while (input.size() > 0)
    {
        StdioParser parser(input);
        
        // if the chunk is not yet complete, try again later
        if (!parser.parse())
        {
            W("contents invalid");
            return;
        }
        
        input = parser.getLeftBytes();
        int commandNumber = parser.getCommandNumber();
        int returnCode    = parser.getErrorCode();
        
        // TODO: we could theoretically inform the calling process that
        // parts of the data could have been queried already
        if (output.contains(parser.getCommandNumber()))
        {
            output[commandNumber].append(parser.getPayload());
        }
        else
        {
            output.insert(commandNumber, parser.getPayload());
        }
        
        // check if this was the last output
        if (parser.getChunkType() == 'l')
        {
            // remember the return value of the command 
            completedCommands.insert(commandNumber, returnCode);
            D(QString("command %1 finished").arg(commandNumber));
            emit commandFinished();
        }
    }
    
    // if this was not the last output, we need to wait for more data
	// FIXME: this is wrong. There is no return in front of this code ATM.
    // W("Contents incomplete.");
    return;
}

QByteArray Monotone::getRawData(int commandNumber)
{
    QMutexLocker locker(&lock);
    
    if (!output.contains(commandNumber))
    {
        W(QString("no data for %1").arg(commandNumber));
        return QByteArray();
    }
    
    QByteArray temp(output.value(commandNumber));
    output.remove(commandNumber);
    return temp;
}

QString Monotone::getDecodedData(int commandNumber)
{
    return QString::fromUtf8(getRawData(commandNumber).data());
}

int Monotone::getReturnCode(int commandNumber)
{
    QMutexLocker locker(&lock);
    
    if (!completedCommands.contains(commandNumber))
    {
        W(QString("no return code for %1").arg(commandNumber));
        return -1;
    }
    
    int returnCode = completedCommands.value(commandNumber);
    // FIXME: shall we really remove this one here? I.e. isCommandFinished()
    // will return false if it is called afterwards
    completedCommands.remove(commandNumber);
    return returnCode;
}

bool Monotone::isCommandFinished(int commandNumber) const
{
    return completedCommands.contains(commandNumber);
}

bool Monotone::runCommand(const QString & path, const QStringList & params, QString & output)
{
    QProcess proc;
    proc.start(path, params);
    
    // could not be started (invalid path, not executable, ...)
    if (!proc.waitForStarted(5000))
    {
        D("process not started");
        return false;
    }
    
    // process doesn't return, etc...
    if (!proc.waitForFinished(5000))
    {
        D("process not finished");
        return false;
    }
    
    // read all of the output
    output.append(proc.readAll());
    return true;
}
    
bool Monotone::checkProgramVersion(const QString & path)
{
    QString output;
    QStringList opts;
    opts << "--version";
    
    if (!runCommand(path, opts, output))
    {
        return false;
    }
    
    QRegExp regex("^monotone (\\d+(?:\\.\\d+))");
    
    if (regex.indexIn(output) == -1)
    {
        D(QString("couldn't parse output: %1").arg(output));
        return false;
    }
    
    QString curVersion  = regex.cap(1);
    
    D(QString("monotone version: %1, need: %2")
        .arg(curVersion)
        .arg(RequiredProgramVersion)
    );
    
    return versionCompare(curVersion, RequiredProgramVersion) >= 0;
}

bool Monotone::checkInterfaceVersion(const QString & path)
{
    QString output;
    QStringList opts;
    opts << "automate" << "interface_version";
    
    if (!runCommand(path, opts, output))
    {
        return false;
    }
    
    QRegExp regex("^(\\d+(?:\\.\\d+))");
    
    if (regex.indexIn(output) == -1)
    {
        D(QString("couldn't parse output: %1").arg(output));
        return false;
    }
    
    QString curVersion  = regex.cap(1);
    QStringList needVersion = RequiredInterfaceVersion.split(".");
    
    D(QString("interface version: %1, need: %2")
        .arg(curVersion)
        .arg(RequiredInterfaceVersion)
    ); 

    return versionCompare(curVersion, RequiredInterfaceVersion) >= 0;
}

// strip mtn: warning: and alike from error strings coming from mtn
// and unwrap them, if this fails we add the original string unwrapped
QString Monotone::stripMtnPrefix(const QString & input)
{
    QString output;
    QStringList list = input.split(QRegExp("\\n"));
    
    // FIXME: we should actually use basename(mtnBinaryPath) in case
    // the binary is renamed, but I'm too lazy for this right now...
    // and this would also require to make this method non-static
    QRegExp rx = QRegExp("^(?:mtn\\:[ ]+)?[^:]+\\:[ ]*(.+)");
    for (int i=0, j=list.size(); i<j; i++)
    {
        if (rx.indexIn(list[i]) != -1)
        {
            output += rx.cap(1) + " ";
            continue;
        }
        output += list[i] + " ";
    }
    return output.trimmed();
}

// Returns
// -1 if right is greater
//  0 if both are equal
//  1 if left is greater
// FIXME: this should be moved somewhere else
int Monotone::versionCompare(const QString & left, const QString & right)
{
    QStringList leftParts = left.split(".");
    QStringList rightParts = right.split(".");
    
    int leftCount = leftParts.size();
    int rightCount = rightParts.size();
    int maxCount = leftCount > rightCount ? leftCount : rightCount;
    
    for (int i=0, j=maxCount; i<j; i++)
    {
        unsigned int l = 0;
        if (i < leftCount) { l = leftParts.at(i).toUInt(); }
        unsigned int r = 0;
        if (i < rightCount) { r = rightParts.at(i).toUInt(); }
        
        if (l == r) continue;
        if (l > r) return 1;
        return -1;
    }
    
    return 0;
}

// convert QStrings to their UTF8 byte representation
ByteArrayList Monotone::stringToByteArrayList(const QStringList & list)
{
    ByteArrayList byteArrayList;
    foreach (QString entry, list)
    {
        byteArrayList.append(entry.toUtf8());
    }
    return byteArrayList;
}

