/***************************************************************************
*   Copyright (C) 2007 by Thomas Keller                                   *
*   me@thomaskeller.biz                                                   *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
***************************************************************************/

#include "MonotoneDelegate.h"
#include "Monotone.h"
#include "Guitone.h"
#include "BasicIOParser.h"

MonotoneDelegate::MonotoneDelegate(AutomateCommand * cmd) 
    : cmdModel(cmd), commandNumber(-1)
{
}

MonotoneDelegate::~MonotoneDelegate() {}

bool MonotoneDelegate::triggerCommand(const QStringList & cmd)
{
    return triggerCommand(cmd, QStringList());
}

bool MonotoneDelegate::triggerCommand(const QStringList & cmd, const QStringList & opts)
{
    QObject * obj = dynamic_cast<QObject*>(cmdModel);
    I(obj);
    Monotone * mtn = MTN(obj);

    connect(
        mtn, SIGNAL(commandFinished()),
        this, SLOT(commandFinished())
    );

    APP->setOverrideCursor(Qt::WaitCursor);

    return mtn->triggerCommand(cmd, opts, commandNumber);
}

// FIXME: if a request is aborted, commandFinished is never signalled,
// do we need to take care of that here?
void MonotoneDelegate::commandFinished()
{
    QObject * obj = dynamic_cast<QObject*>(cmdModel);
    I(obj);
    Monotone * mtn = MTN(obj);
    
    // check if this is the command which is interesting for us
    if (!mtn->isCommandFinished(commandNumber)) return;
    
    disconnect(
       mtn, SIGNAL(commandFinished()),
       this, SLOT(commandFinished())
    );

    // FIXME: does any of our models ever need the raw data?
    cmdModel->setAutomateData(mtn->getDecodedData(commandNumber));
    int returnCode = mtn->getReturnCode(commandNumber);

    if (returnCode == 0)
    {
       cmdModel->parseOutput();
    }
    else
    {
        if (!cmdModel->handleError(returnCode))
        {
            C(QString("Couldn't handle error %1").arg(returnCode));
        }
    }

    APP->restoreOverrideCursor();
}

QString MonotoneDelegate::getBaseWorkspaceRevision(QObject * obj)
{
    Monotone * mtn = MTN(obj);
    
    int commandNumber;
    if (!mtn->executeCommand(QStringList() << "get_base_revision_id", commandNumber) ||
        mtn->getReturnCode(commandNumber) > 0)
    {
        W("Could not execute get_base_revision_id");
        return false;
    }
    
    QString baseRevision = mtn->getDecodedData(commandNumber);
    baseRevision.chop(1);
    return baseRevision;
}

QString MonotoneDelegate::getOption(QObject * obj, const QString & opt)
{
    Monotone * mtn = MTN(obj);
    
    int cmdNum;
    mtn->executeCommand(QStringList() << "get_option" << opt, cmdNum);
    
    QString data = mtn->getDecodedData(cmdNum);
    
    if (mtn->getReturnCode(cmdNum) > 0)
    {
        C(QString("Couldn't retrieve option %1: %2").arg(opt).arg(data));
        return QString();
    }
    
    // remove the trailing newline
    data.chop(1);
    
    return data;
}

QString MonotoneDelegate::getBranchName(QObject * obj, const QString & defaultBranch)
{
    QString branchName = getOption(obj, "branch");
    
    if (branchName.size() == 0)
    {
        branchName = defaultBranch;
    }
    return branchName;
}

QString MonotoneDelegate::getBranchNameShort(QObject * obj)
{
    // FIXME: currently there is only support for the domain branch scheme
    // tld.domain.project.branch
    QString branchName = getBranchName(obj);
    QStringList parts = branchName.split('.');
    if (parts.size() == 1) return branchName;
    
    int maxSize = 30;
    int minNeededChars = (parts.size() * 2) - 1;
    
    QString shortBranchName;
    for (int i=parts.size()-1; i>=0 ; i--)
    {
        QString part = parts.at(i);
        if ((part.size() + shortBranchName.size() + minNeededChars) > maxSize)
        {
            shortBranchName.prepend(part.at(0));
        }
        else
        {
            shortBranchName.prepend(part);
        }
        
        if (i != 0) shortBranchName.prepend(".");
        minNeededChars -= 2;
    }
    
    return shortBranchName;
}

QStringList MonotoneDelegate::resolveSelector(QObject * obj, const QString & selector)
{
    Monotone * mtn = MTN(obj);
    
    int cmdNum;
    mtn->executeCommand(QStringList() << "select" << selector, cmdNum);
    
    QString data = mtn->getDecodedData(cmdNum);
    
    if (mtn->getReturnCode(cmdNum) > 0)
    {
        C(QString("Couldn't resolve selector %1: %2").arg(selector).arg(data));
        return QStringList();
    }
    
    QStringList revList = data.split('\n', QString::SkipEmptyParts);
    return revList;
}

RevisionCerts MonotoneDelegate::getRevisionCerts(QObject * obj, const QString & revision)
{
    Monotone * mtn = MTN(obj);
    
    int cmdNum;
    mtn->executeCommand(QStringList() << "certs" << revision, cmdNum);
    
    QString data = mtn->getDecodedData(cmdNum);
    RevisionCerts certs;
    
    if (mtn->getReturnCode(cmdNum) > 0)
    {
        C(QString("Couldn't query revision certs for %1: %2").arg(revision).arg(data));
        return certs;
    }
    
    BasicIOParser parser(data);
    if (!parser.parse())
    {
        C("Could not parse basic_io.");
        return certs;
    }
    
    StanzaList stanzas = parser.getStanzas();
    foreach (Stanza st, stanzas)
    {
        RevisionCert cert;
        
        foreach (StanzaEntry en, st)
        {
            if (en.sym == "name")
            {
                cert.first = en.vals.at(0);
            }
            
            if (en.sym == "value")
            {
                cert.second = en.vals.at(0);
            }
        }
        I(!cert.first.isEmpty() && !cert.second.isEmpty());
        certs.append(cert);
    }
    return certs;
}

QString MonotoneDelegate::getDatabaseFilePath(QObject * obj)
{
    Monotone * mtn = MTN(obj);
    QString path = mtn->getDatabaseFilePath();
    
    // apparently a workspace is loaded...
    if (path.isEmpty())
    {
        path = getOption(obj, "database");
    }
    
    I(!path.isEmpty());
    return path;
}

FileEntryList MonotoneDelegate::getRevisionManifest(QObject * obj, const QString & revision)
{
    Monotone * mtn = MTN(obj);
    
    int cmdNum;
    mtn->executeCommand(QStringList() << "get_manifest_of" << revision, cmdNum);
    
    QString data = mtn->getDecodedData(cmdNum);
    FileEntryList entries;
    
    if (mtn->getReturnCode(cmdNum) > 0)
    {
        C(QString("Couldn't query manifest entries for %1: %2").arg(revision).arg(data));
        return entries;
    }
    
    BasicIOParser parser(data);
    if (!parser.parse())
    {
        C("Could not parse basic_io.");
        return entries;
    }
    
    StanzaList stanzas = parser.getStanzas();
    foreach (Stanza st, stanzas)
    {
        foreach (StanzaEntry en, st)
        {
            if (en.sym == "format_version") break;
        
            FileEntry entry;
            entry.first = en.vals.at(0);
            I(en.sym == "dir" || en.sym == "file");
            entry.second = en.sym == "dir";
         
            entries.append(entry);
            break;
        }
    }
    return entries;
}

QStringList MonotoneDelegate::getPrivateKeyList(QObject * obj)
{
    Monotone * mtn = MTN(obj);
    
    int cmdNum;
    mtn->executeCommand(QStringList() << "keys", cmdNum);
    
    QString data = mtn->getDecodedData(cmdNum);
    QStringList keys;
    
    if (mtn->getReturnCode(cmdNum) > 0)
    {
        C(QString("Couldn't query keys: %1").arg(data));
        return keys;
    }
    
    BasicIOParser parser(data);
    if (!parser.parse())
    {
        C("Could not parse basic_io.");
        return keys;
    }
    
    StanzaList stanzas = parser.getStanzas();
    foreach (Stanza st, stanzas)
    {
        QString lastKey;
        foreach (StanzaEntry en, st)
        {
            if (en.sym == "name")
            {
                lastKey = en.vals.at(0);
                continue;
            }
            
            if (en.sym == "private_location")
            {
                I(!lastKey.isEmpty());
                keys.append(lastKey);
                break;
            }
        }
    }
    return keys;
}
