/***************************************************************************
 *   Copyright (C) 2007 by Thomas Keller                                   *
 *   me@thomaskeller.biz                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
 
#include "WorkspaceCommitter.h"
#include "MonotoneDelegate.h"
#include "Guitone.h"
#include "BasicIOParser.h"

#include <QProgressDialog>
#include <QDateTime>

WorkspaceCommitter::WorkspaceCommitter(QObject * parent, QString k, QString b, QString a, QString c)
    : QObject(parent), key(k)
{
    // use the signing key as default author cert value
    if (a.isEmpty()) a = k;
    
    certs.insert("branch", b);
    certs.insert("author", a);
    certs.insert("changelog", c);
    certs.insert("date", 
                 QDateTime::currentDateTime().toUTC().toString(Qt::ISODate));
    
    workspaceDir = QDir(MTN(this)->getNormalizedWorkspacePath());
    I(workspaceDir.exists());
}

WorkspaceCommitter::~WorkspaceCommitter()
{
}

bool WorkspaceCommitter::run()
{
    // at first get the current revision text
    // FIXME: we don't yet check here if this has change since we openend the
    // commit dialog!
    Monotone * mtn = MTN(this);
    
    int cmdNum;
    mtn->executeCommand(QStringList() << "get_revision", cmdNum);
    
    QString data = mtn->getDecodedData(cmdNum);
    
    if (mtn->getReturnCode(cmdNum) > 0)
    {
        C(QString("Couldn't query workspace revision: %1").arg(data));
        return false;
    }
    
    QString fullRevision = data;
    
    BasicIOParser parser(fullRevision);
    if (!parser.parse())
    {
        C("Could not parse basic_io.");
        return false;
    }
    
    QList<QPair<QString, QString> > fileChanges;
    
    StanzaList stanzas = parser.getStanzas();
    foreach (Stanza st, stanzas)
    {
        QPair<QString, QString> pair;
        bool isPatch = false;
        
        foreach (StanzaEntry en, st)
        {
            if (en.sym == "add_file")
            {
                pair.first = en.vals.at(0);
                fileChanges.append(pair);
                break;
            }
            
            if (en.sym == "patch")
            {
                pair.first = en.vals.at(0);
                isPatch = true;
                continue;
            }
            
            if (isPatch && en.sym == "from")
            {
                I(!pair.first.isEmpty());
                pair.second = en.vals.at(0);
                fileChanges.append(pair);
                break;
            }
        }
    }
    
    // put files
    for (int i=0, j=fileChanges.size(); i < j; i++)
    {
        QPair<QString, QString> pair = fileChanges.at(i);
        if (!putFile(pair.first, pair.second))
        {
            // FIXME: what to do with already added file deltas?
            return false;
        }
    }
    
    // commit revision
    mtn->executeCommand(QStringList() << "put_revision" << fullRevision, cmdNum);
    
    data = mtn->getDecodedData(cmdNum);
    
    if (mtn->getReturnCode(cmdNum) > 0)
    {
        C(QString("Couldn't commit revision: %1").arg(data));
        return false;
    }
    
    data.chop(1);
    revisionId = data;
    
    // attach certificates
    QMapIterator<QString, QString> i(certs);
    while (i.hasNext())
    {
        i.next();
        
        int cmdNum;
        mtn->executeCommand(
            QStringList() << "cert" << revisionId << i.key() << i.value(),
            QStringList() << "key" << key,
            cmdNum
        );
        
        QString data = mtn->getDecodedData(cmdNum);
        
        if (mtn->getReturnCode(cmdNum) > 0)
        {
            C(QString("Couldn't attach cert %1: %2").arg(i.key()).arg(data));
            return false;
        }
    }
    
    D(QString("Committed revision %1").arg(revisionId));
    return true;
}

bool WorkspaceCommitter::putFile(const QString & fileName, const QString & baseId)
{
    QFile file(workspaceDir.filePath(fileName));
    if (!file.exists())
    {
        // FIXME: we need to bubble this up to the user, this typically
        // happens if there are missing files in the workspace
        C(QString("File %1 does not exist in workpace.").arg(fileName));
        return false;
    }
    
    if (!file.open(QIODevice::ReadOnly))
    {
        C("Can't open file for reading.");
        return false;
    }
    
    Monotone * mtn = MTN(this);
    
    int commandNumber;
    
    ByteArrayList command;
    command << "put_file";
    if (!baseId.isEmpty())
    {
        command << baseId.toAscii();
    }
    command << file.readAll();
    file.close();
    
    if (!mtn->executeCommand(command, commandNumber) ||
        mtn->getReturnCode(commandNumber) > 0)
    {
        C(QString("Cannot execute put_file: %1")
            .arg(mtn->getDecodedData(commandNumber)));
        return false;
    }
    
    return true;
}

