/***************************************************************************
 *   Copyright (C) 2006 by Thomas Keller                                   *
 *   me@thomaskeller.biz                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "InventoryView.h"
#include "Settings.h"
#include "Platform.h"
#include "InventoryItem.h"
#include "Monotone.h"
#include "FileDiff.h"
#include "RevisionDiff.h"
#include "vocab.h"
#include "Guitone.h"

#include <QMenu>
#include <QHeaderView>
#include <QSortFilterProxyModel>
#include <QFileInfo>
#include <QMessageBox>
#include <QTimer>

InventoryView::InventoryView(QWidget* parent)
: TreeView(parent), invViewDelegate(parent)
{
    setItemDelegate(&invViewDelegate);
    setSelectionMode(QAbstractItemView::ExtendedSelection);
    setDragEnabled(true);
    setDropIndicatorShown(true);
    setSortingEnabled(true);
    sortByColumn(0, Qt::AscendingOrder);
    
    header()->setClickable(true);
    
    createAndConnectContextActions();
    
    connect(
        this, SIGNAL(doubleClicked(const QModelIndex &)),
        this, SLOT(itemClicked(const QModelIndex &))
    );
    
    connect(
        this, SIGNAL(contextMenuRequested(const QModelIndexList &, const QPoint &)),
        this, SLOT(slotContextMenuRequested(const QModelIndexList &, const QPoint &))
    );
}

void InventoryView::setType(Type t)
{
    // ensure that a model is already set, otherwise the app crashes
    I(model());
    
    type = t;
    
    if (type == FileList)
    {
        disconnect(
            this, SIGNAL(clicked(const QModelIndex &)),
            this, SLOT(itemClicked(const QModelIndex &))
        );
     
        setRootIsDecorated(false);
        setItemsExpandable(false);
    }
    else
    {
        connect(
            this, SIGNAL(clicked(const QModelIndex &)),
            this, SLOT(itemClicked(const QModelIndex &))
        );
        setRootIsDecorated(true);
        setItemsExpandable(true);
    }
}

InventoryView::~InventoryView()
{
    delete actChdir;
    delete actOpen; 
    delete actAdd; 
    delete actRemove; 
    delete actCommit; 
    delete actIgnore; 
    delete actUnignore; 
    delete actRevert; 
    delete actFileDiff; 
    delete actRevisionDiff; 
    delete actRename;
    delete actAddMultiple;
    delete actRemoveMultiple;
    delete actCommitMultiple;
    delete actIgnoreMultiple;
    delete actUnignoreMultiple;
    delete actRevertMultiple;
}

// ensure that we only accept sortfilterproxymodels
void InventoryView::setModel(QAbstractItemModel * model)
{
    I(false);
}

void InventoryView::setModel(QSortFilterProxyModel * newModel)
{
    QAbstractItemModel * oldModel = model();
    if (oldModel)
    {
        disconnect(
            oldModel, SIGNAL(modelReset()),
            this, SLOT(delegateModelReset())
        );
    }
    
    TreeView::setModel(newModel);

    connect(
        newModel, SIGNAL(modelReset()),
        this, SLOT(delegateModelReset())
    );
}

void InventoryView::delegateModelReset()
{
    // give the view another event cycle to reload the new indices
    //QTimer::singleShot(0, this, SLOT(modelReset()));
    modelReset();
}

void InventoryView::modelReset()
{
    QModelIndex index = model()->index(0, 0, QModelIndex());
    if (!index.isValid())
    {
        C("Index invalid, aborting...");
        return;
    }
    
    if (type == FileList)
    {
        setRootIndex(index);
    }
    else
    {
        expand(index);
    }
}

// adds elements to the popup menu based on the selected items
// since not all actions may apply on all items
void InventoryView::slotContextMenuRequested(const QModelIndexList & indexList, const QPoint & pos)
{
    D(QString("Selected indexes: %1").arg(indexList.size()));
    
    QMenu menu(this);
    
    if (indexList.size() == 1)
    {
        QModelIndex sourceIndex = static_cast<const QSortFilterProxyModel*>(indexList[0].model())->mapToSource(indexList[0]);
        
        InventoryItem* item = static_cast<InventoryItem*>(sourceIndex.internalPointer());
        
        
        if (item->isDirectory())
        {
            menu.addAction(actChdir);
        }
        
        menu.addAction(actOpen);
        
        if (item->hasStatus(InventoryItem::Unknown))
        {
            menu.addAction(actAdd);
            menu.addAction(actIgnore);
        }
        
        if (item->hasStatus(InventoryItem::Ignored))
        {
            menu.addAction(actUnignore);
        }
        
        if (item->hasChangedRecursive())
        {
            menu.addAction(actCommit);
            menu.addAction(actRevisionDiff);
        }
        
        if (item->hasNotStatus(InventoryItem::Ignored) && 
            item->hasNotStatus(InventoryItem::Unknown) &&
            !item->isRootDirectory() && !item->isCdUp())
        {
            menu.addAction(actRemove);
            menu.addAction(actRename);
            
            // as long as it is not unchanged, it can be converted
            if (item->hasNotStatus(InventoryItem::Unchanged))
            {
                menu.addAction(actRevert);
                
                if (item->hasNotStatus(InventoryItem::Added) &&
                    !item->isDirectory())
                {
                    menu.addAction(actFileDiff);
                }   
            }
        }
        
        //
        // determine the default action
        //
        QFont activeFont;
        activeFont.setBold(true);
        QFont normalFont;
        normalFont.setBold(false);
        
        if (item->isDirectory())
        {
            actOpen->setFont(normalFont);
            actChdir->setFont(activeFont);
        }
        else
        {
            if (item->hasStatus(InventoryItem::Patched) && 
                item->hasNotStatus(InventoryItem::Added))
            {
                actFileDiff->setFont(activeFont);
                actOpen->setFont(normalFont);
            }
            else
            {
                actFileDiff->setFont(normalFont);
                actOpen->setFont(activeFont);
            }
        }
        
        menu.exec(pos);
    }
    else
    {
        InventoryItem * item;
        QModelIndex sourceIndex;
        
        QStringList actions;
        actions << "add" << "drop" << "commit" 
                << "ignore" << "unignore" << "revert";
                
        QMap<QString, QAction*> actionMapping;
        actionMapping.insert("add", actAddMultiple);
        actionMapping.insert("drop", actRemoveMultiple);
        actionMapping.insert("commit", actCommitMultiple);
        actionMapping.insert("revert", actRevertMultiple);
        actionMapping.insert("ignore", actIgnoreMultiple);
        actionMapping.insert("unignore", actUnignoreMultiple);
        
        QMap<QString, int> actionCounter;
        
        foreach(QString action, actions)
        {
            actionCounter.insert(action, 0);
        }
        
        foreach(QModelIndex index, indexList)
        {
            sourceIndex = 
                static_cast<const QSortFilterProxyModel*>(index.model())
                    ->mapToSource(index);

            item = static_cast<InventoryItem*>(sourceIndex.internalPointer());
            
            if (!item->isTracked())
            {
                actionCounter["add"]++;
                
                if (item->hasStatus(InventoryItem::Unknown))
                {
                    actionCounter["ignore"]++;
                }
                else
                {
                    actionCounter["unignore"]++;
                }
                continue;
            }
            
            if (item->hasChangedRecursive())
            {
                actionCounter["commit"]++;
                actionCounter["revert"]++;
            }
            
            actionCounter["drop"]++;
        }
        
        QMapIterator<QString, int> i(actionCounter);
        int totalItems = indexList.size();
        bool commonActions = false;
        
        while (i.hasNext())
        {
            i.next();
            if (i.value() < totalItems) continue;
            commonActions = true;
            QAction * act = actionMapping[i.key()];
            act->setText(act->text().arg(totalItems));
            menu.addAction(act);
        }
        
        if (!commonActions)
        {
            QAction * act = new QAction(
                tr("No common action for selected items"), this
            );
            menu.addAction(act);
        }
        
        menu.exec(pos);
        return;
    }
}

void InventoryView::createAndConnectContextActions(void)
{
    actChdir = new QAction(tr("Go into"), this); 
    actChdir->setStatusTip(tr("Go into the directory"));
    connect(actChdir, SIGNAL(triggered()), this, SLOT(slotChdir()));

    actOpen = new QAction(tr("Open"), this); 
    actOpen->setStatusTip(tr("Open in default program")); 
    connect(actOpen, SIGNAL(triggered()), this, SLOT(slotOpen()));

    actAdd = new QAction(tr("Add"), this); 
    actAdd->setStatusTip(tr("Add to workspace")); 
    connect(actAdd, SIGNAL(triggered()), this, SLOT(slotAdd()));

    actRemove = new QAction(tr("Remove"), this); 
    actRemove->setStatusTip(tr("Remove from workspace")); 
    connect(actRemove, SIGNAL(triggered()), this, SLOT(slotRemove()));

    actCommit = new QAction(tr("Commit"), this); 
    actCommit->setStatusTip(tr("Commit"));
    connect(actCommit, SIGNAL(triggered()), this, SLOT(slotCommit()));

    actIgnore = new QAction(tr("Ignore"), this); 
    actIgnore->setStatusTip(tr("Ignore file"));
    connect(actIgnore, SIGNAL(triggered()), this, SLOT(slotIgnore()));
    
    actUnignore = new QAction(tr("Unignore"), this); 
    actUnignore->setStatusTip(tr("Unignore file"));
    connect(actUnignore, SIGNAL(triggered()), this, SLOT(slotUnignore()));
    
    actRevert = new QAction(tr("Revert"), this); 
    actRevert->setStatusTip(tr("Revert uncommitted changes"));
    connect(actRevert, SIGNAL(triggered()), this, SLOT(slotRevert()));
    
    actFileDiff = new QAction(tr("Diff"), this); 
    actFileDiff->setStatusTip(tr("Diff against base revision"));
    connect(actFileDiff, SIGNAL(triggered()), this, SLOT(slotFileDiff()));
  
    actRevisionDiff = new QAction(tr("Diff all"), this); 
    actRevisionDiff->setStatusTip(tr("Show all differences"));
    connect(actRevisionDiff, SIGNAL(triggered()), this, SLOT(slotRevisionDiff()));
    
    actRename = new QAction(tr("Rename"), this); 
    actRename->setStatusTip(tr("Rename file"));
    connect(actRename, SIGNAL(triggered()), this, SLOT(slotRename()));
    
    actAddMultiple = new QAction(tr("Add %1 items"), this);
    actAddMultiple->setStatusTip(tr("Add multiple items"));
    connect(actAddMultiple, SIGNAL(triggered()), this, SLOT(slotAdd()));
   
    actRemoveMultiple = new QAction(tr("Remove %1 items"), this);
    actRemoveMultiple->setStatusTip(tr("Remove multiple items"));
    connect(actRemoveMultiple, SIGNAL(triggered()), this, SLOT(slotRemove()));
    
    actCommitMultiple = new QAction(tr("Commit %1 items"), this);
    actCommitMultiple->setStatusTip(tr("Commit multiple items"));
    connect(actCommitMultiple, SIGNAL(triggered()), this, SLOT(slotCommit()));
    
    actIgnoreMultiple = new QAction(tr("Ignore %1 items"), this);
    actIgnoreMultiple->setStatusTip(tr("Ignore multiple items"));
    connect(actIgnoreMultiple, SIGNAL(triggered()), this, SLOT(slotIgnore()));
   
    actUnignoreMultiple = new QAction(tr("Unignore %1 items"), this);
    actUnignoreMultiple->setStatusTip(tr("Unignore multiple items"));
    connect(actUnignoreMultiple, SIGNAL(triggered()), this, SLOT(slotUnignore()));
   
    actRevertMultiple = new QAction(tr("Revert %1 items"), this);
    actRevertMultiple->setStatusTip(tr("Revert multiple items"));
    connect(actRevertMultiple, SIGNAL(triggered()), this, SLOT(slotRevert()));
}

void InventoryView::slotChdir()
{
    QItemSelectionModel *selectionModel = this->selectionModel();
    QList<QModelIndex> list(selectionModel->selectedIndexes());
    I(list.size() > 0);
    
    emit directoryChanged(list[0]);
    changeDirectory(list[0]);
}

void InventoryView::changeDirectory(const QModelIndex & proxyIndex)
{
    QModelIndex source = static_cast<QSortFilterProxyModel*>(model())->mapToSource(proxyIndex);
    InventoryItem *item = static_cast<InventoryItem*>(source.internalPointer());
    
    if (type == FileList)
    {
        // retranslate the proxy index to an index of the current model
        QModelIndex proxyFileIndex = 
            static_cast<QSortFilterProxyModel*>(model())->mapFromSource(source);

        if (item->isCdUp())
        {
            // list the contents of the parent of the parent directory
            setRootIndex(proxyFileIndex.parent().parent());
            return;
        }
        setRootIndex(proxyFileIndex);
        return;
    }
    
    // retranslate the proxy index to an index of the current model
    QModelIndex proxyFolderIndex = 
        static_cast<QSortFilterProxyModel*>(model())->mapFromSource(source);
        
    // expand the selection if needed
    expand(proxyFolderIndex);
    
    // select the item
    selectionModel()->setCurrentIndex(
        proxyFolderIndex,
        QItemSelectionModel::ClearAndSelect | QItemSelectionModel::Rows
    );
}

void InventoryView::slotOpen(void)
{
    QModelIndex index(getSingleSelection());
    if (!index.isValid()) return;
    
    InventoryItem * item = static_cast<InventoryItem*>(index.internalPointer());
    
    QFileInfo file(MTN(this)->getNormalizedWorkspacePath() + "/" + item->getPath());
    if (!file.exists())
    {
        
        QMessageBox::critical(
            this,
            tr("Error"),
            tr("The file you're trying to open does not exist."),
            QMessageBox::Ok, 0, 0
        );
        return;
    }
    
    if (!Platform::openFile(this, file.absoluteFilePath()))
    {
        QMessageBox::critical(
            this,
            tr("Error"),
            tr("Unable to open files on your platform - please contact the "
            "author about this problem."),
            QMessageBox::Ok, 0, 0
        );
    }
}

void InventoryView::slotAdd(void)
{
    C("Not implemented.");
}

void InventoryView::slotRemove(void)
{
    C("Not implemented.");
}

void InventoryView::slotCommit(void)
{
    C("Not implemented.");
}

void InventoryView::slotRevert(void)
{
    C("Not implemented.");
}

void InventoryView::slotRename(void)
{
    C("Not implemented.");
}

//
// TODO: there is no mtn API for the next two commands available 
// right now so we need to manually edit the WD/.mtn-ignore file
// Q: If this file is under version control, should a change
// be committed in the background?
//
void InventoryView::slotIgnore(void)
{
    C("Not implemented.");
}

void InventoryView::slotUnignore(void)
{
    C("Not implemented.");
}

void InventoryView::slotFileDiff(void)
{
    QModelIndex index(getSingleSelection());
    if (!index.isValid()) return;
    
    InventoryItem * item = static_cast<InventoryItem*>(index.internalPointer());
    
    if (item->isDirectory() || !item->hasStatus(InventoryItem::Patched)) 
    {
        D("File is a directory or not patched/unknown.");
        return;
    }
    
    FileDiff dlg(this, item->getPath());
    dlg.exec();

    clearSelection();
}

void InventoryView::slotRevisionDiff(void)
{
    QModelIndex index(getSingleSelection());
    if (!index.isValid()) return;
    
    InventoryItem * item = static_cast<InventoryItem*>(index.internalPointer());
    
    RevisionDiff dlg(this);
    dlg.init(item->getPath(), QString(), QString());
    dlg.exec();

    clearSelection();
}

QModelIndex InventoryView::getSingleSelection(bool mapToSource /* = true */)
{
    QItemSelectionModel * selectionModel = this->selectionModel();
    QList<QModelIndex> list(selectionModel->selectedIndexes());
    
    if (list.size() == 0)
    {
        D("No item selected.");
        return QModelIndex();
    }
    
    if (list.size() > 1)
    {
        D("Multiple items selected, only returning the first.");
    }
    
    if (!mapToSource) return list[0];
    
    return static_cast<const QSortFilterProxyModel*>(list[0].model())->mapToSource(list[0]);    
}

void InventoryView::itemClicked(const QModelIndex & index)
{
    if (!index.isValid()) return;
    QModelIndex sourceIndex = 
        static_cast<const QSortFilterProxyModel*>(index.model())->mapToSource(index);
    
    InventoryItem * item = static_cast<InventoryItem*>(sourceIndex.internalPointer());
    
    if (item->isDirectory())
    {
        emit directoryChanged(index);
        changeDirectory(index);
        return;
    }
    
    if (item->hasStatus(InventoryItem::Patched) &&
        item->hasNotStatus(InventoryItem::Added))
    {
        slotFileDiff();
        return;
    }
    
    // default action on double click
    slotOpen();
    return;
}

