/***************************************************************************
 *   Copyright (C) 2006 by Thomas Keller                                   *
 *   me@thomaskeller.biz                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "MainWindow.h"
#include "Monotone.h"
#include "Inventory.h"
#include "InventoryItem.h"
#include "Attributes.h"
#include "InventoryProxyModel.h"
#include "Splitter.h"
#include "InventoryView.h"
#include "AttributesView.h"
#include "UpdateWorkspace.h"
#include "CommitRevision.h"
#include "CheckoutRevision.h"
#include "ApplicationUpdate.h"
#include "Platform.h"
#include "Preferences.h"
#include "KeyManagement.h"
#include "About.h"
#include "Settings.h"
#include "ChangesetBrowser.h"
#include "WorkspaceCreator.h"
#include "Guitone.h"

#ifdef Q_WS_MAC
#include "CocoaUtil.h"
#endif

#include <QMessageBox>
#include <QStatusBar>
#include <QMenuBar>
#include <QFileDialog>
#include <QCloseEvent>
#include <QFileInfo>
#include <QDesktopWidget>

MainWindow::MainWindow()
    : QMainWindow(), closeCounter(0)
{
    setAttribute(Qt::WA_MacMetalStyle);

    // ensure that the shortcut keys are properly recognized by linguist
    QShortcut::tr("Ctrl");
    QShortcut::tr("Alt");
    QShortcut::tr("Shift");
    QShortcut::tr("Meta");
    
    setupUi(this);
    
    // create the main models
    invModel = new Inventory(this);
    attrModel = new Attributes(this);
    
    connect(
        invModel, SIGNAL(invalidWorkspaceFormat(const QString &)),
        this, SLOT(invalidWorkspaceFormat(const QString &))
    );
    
    // ProxyModels
    proxyModelFolderTree = new InventoryProxyModel(this, true);
    proxyModelFileList = new InventoryProxyModel(this, false);
    proxyModelFolderTree->setSourceModel(invModel);
    proxyModelFileList->setSourceModel(invModel);

    // tree and list views
    treeView->setModel(proxyModelFolderTree);
    treeView->setType(InventoryView::FolderTree);
    listView->setModel(proxyModelFileList);
    listView->setType(InventoryView::FileList);
    attrView->setModel(attrModel);
    
    // query attributes on click
    connect(
        treeView, SIGNAL(clicked(const QModelIndex &)),
        attrModel, SLOT(readAttributes(const QModelIndex &))
    );
    connect(
        listView, SIGNAL(clicked(const QModelIndex &)),
        attrModel, SLOT(readAttributes(const QModelIndex &))
    );
    
    // filelist/tree synchronization
    connect(
        treeView, SIGNAL(directoryChanged(const QModelIndex &)),
        listView, SLOT(changeDirectory(const QModelIndex &))
    );
    connect(
        listView, SIGNAL(directoryChanged(const QModelIndex &)),
        treeView, SLOT(changeDirectory(const QModelIndex &))
    );

    // rename the appQuit signal
    connect(
        actionQuit, SIGNAL(triggered()),
        this, SIGNAL(quitApplication())
    );
    
    // update recent workspace and database lists on request
	connect(
        this, SIGNAL(updatePreviousWorkspacesMenu()),
        this, SLOT(doUpdatePreviousWorkspacesMenu())
    );
    
    connect(
        this, SIGNAL(updatePreviousDatabasesMenu()),
        this, SLOT(doUpdatePreviousDatabasesMenu())
    );
    
    doUpdatePreviousWorkspacesMenu();
    doUpdatePreviousDatabasesMenu();
    
    // set the data for some menu elements
    actionAll_files->setData(QVariant(InventoryProxyModel::All));
    actionAll_changed_files->setData(QVariant(InventoryProxyModel::Changed));
    actionPatched_files->setData(QVariant(InventoryProxyModel::Patched));
    actionAdded_files->setData(QVariant(InventoryProxyModel::Added));
    actionRemoved_files->setData(QVariant(InventoryProxyModel::Dropped));
    actionRenamed_files->setData(QVariant(InventoryProxyModel::Renamed));
    actionMissing_files->setData(QVariant(InventoryProxyModel::Missing));
    actionUnknown_files->setData(QVariant(InventoryProxyModel::Unknown));
    actionIgnored_files->setData(QVariant(InventoryProxyModel::Ignored));
    actionExpand_tree->setData(QVariant(false));
    actionHide_ignored_files->setData(QVariant(false));
    
    // after laying out everything, restore the splitter views
    mainSplitter->init();
    listSplitter->init();
}

MainWindow::~MainWindow()
{
    delete invModel;
    delete attrModel;
    delete proxyModelFolderTree;
    delete proxyModelFileList;
}

void MainWindow::on_actionOpen_Workspace_triggered()
{
    QString fn = QFileDialog::getExistingDirectory(0, tr("Select your workspace..."));
    
    if (fn.isEmpty())
    {
        statusBar()->showMessage(tr("Loading aborted"), 2000);
        return;
    }
    
    emit loadWorkspace(fn);	
}

bool MainWindow::doLoadWorkspace(QString fn)
{
    Monotone * mtn = MTN(this);
    
    if (!mtn->loadWorkspace(fn))
    {
        QMessageBox::critical(
            this,
            tr("Failed to load workspace"),
            tr("The workspace could not be loaded.\n"
               "The last output was:\n\n%1").arg(mtn->getStderr()),
            QMessageBox::Ok
        );

        // remove the workspace if it was recorded as recent workspace
        Settings::removeItemFromList("RecentWorkspaceList", fn);
    
        emit updatePreviousWorkspacesMenu();
        
        return false;
    }

    switchMode(Workspace);
    
    if (!invModel->readInventory())
    {
        QMessageBox::information(
            this,
            tr("Unable to execute command"),
            tr("Unable to execute '%1' - maybe another command is still running?").arg("inventory"),
            QMessageBox::Ok
        );
        return false;
    }
    
    // add the workspace to the recent workspace list
    // FIXME: the amount of recent workspaces should be made configurable
    Settings::addItemToList("RecentWorkspaceList", mtn->getNormalizedWorkspacePath(), 5);
    
    emit updatePreviousWorkspacesMenu();
	
    return true;
}

void MainWindow::closeEvent(QCloseEvent * event)
{
    I(closeCounter >= 0);
    
    // ignore the close event if there are still open dialog windows
    if (closeCounter > 0)
    {
        D(QString("Ignoring close request for %1").arg((int)this));
        event->ignore();
        return;
    }
    
    // the last closed window sets the geometry for the next one which is opened
    Settings::setWindowGeometry(saveGeometry(), "MainWindow_mode" + mode);
    event->accept();
    emit windowClosed(this);
}

void MainWindow::on_actionOpen_Database_triggered()
{
    QString fn = QFileDialog::getOpenFileName(
        0,
        tr("Select your database..."),
        QString(),
        tr("monotone Databases (*.mtn *.db)")
    );
    
    if (fn.isEmpty())
    {
        statusBar()->showMessage(tr("Loading aborted"), 2000);
        return;
    }
    
    emit loadDatabase(fn);    
}

bool MainWindow::doLoadDatabase(QString fn)
{
    Monotone * mtn = MTN(this);
    if (!mtn->loadDatabase(fn))
    {
        QMessageBox::critical(
            this,
            tr("Failed to load database"),
            tr("The database could not be loaded.\n"
               "The last output was:\n\n%1").arg(mtn->getStderr()),
            QMessageBox::Ok
        );
        
        Settings::removeItemFromList("RecentDatabaseList", fn);
        
        emit updatePreviousDatabasesMenu();
        
        return false;
    }
    
    switchMode(Database);
    
    Settings::addItemToList("RecentDatabaseList", fn, 5);
	
    emit updatePreviousDatabasesMenu();
    
    return true;
}

void MainWindow::switchMode(Mode m)
{
    mode = m;
    windowMode->setCurrentIndex(mode);
    restoreGeometry(Settings::getWindowGeometry("MainWindow_mode" + mode));
    
    QList<MainWindow*> openWindows = APP->windowList();
    
    // ensure somewhat that new windows do not overdraw current ones
    // by adding a little x/y offset the original position of the window
    // opened before this window
    int curIdx = openWindows.indexOf(this);
    if (curIdx > 0)
    {
        MainWindow * prevWnd = openWindows.at(curIdx - 1);
        I(prevWnd);
        
        QDesktopWidget * desk = APP->desktop();
        
        if (desk->numScreens() > 1)
        {
            W("No support for window cascading on systems with "
              "more than one monitor yet.");
        }
        else
        {
            int cascade = 20;
            QRect geom = desk->availableGeometry();
            int newX = prevWnd->x() + cascade;
            int newY = prevWnd->y() + cascade;
            
            if (newX + width() > geom.right() ||
                newY + height() > geom.bottom())
            {
                newX = geom.x();
                newY = geom.y();
            }
            move(newX, newY);
        }
    }
    
    if (mode == Database)
    {
        menuView->menuAction()->setVisible(false);
        menuWorkspace->menuAction()->setVisible(false);
        menuDatabase->menuAction()->setVisible(true);
        
        QString path(MTN(this)->getDatabaseFilePath());
        
        loadedDatabase->setText(tr("Loaded database: %1").arg(
            MTN(this)->getDatabaseFilePath())
        );
        
        QFileInfo fi(path);
        setWindowTitle(
            tr("%1 - database mode - guitone").arg(fi.fileName())
        );
    }
    else
    if (mode == Workspace)
    {
        menuView->menuAction()->setVisible(true);
        menuWorkspace->menuAction()->setVisible(true);
        menuDatabase->menuAction()->setVisible(true);
        
        setWindowTitle(
            tr("%1 - workspace mode - guitone").
                arg(MonotoneDelegate::getBranchNameShort(this))
        );
    }
    else
    {
        I(false);
    }
    
    emit modeChanged(mode);
}

void MainWindow::on_actionHide_ignored_files_triggered()
{
    bool hide = actionHide_ignored_files->data().toBool();
    proxyModelFolderTree->setHideIgnoredFiles(!hide);
    proxyModelFileList->setHideIgnoredFiles(!hide);
    
    actionHide_ignored_files->setText(
        hide ? tr("Hide ignored files") : tr("Show ignored files")
    );
    
    actionHide_ignored_files->setData(QVariant(!hide));
}

void MainWindow::on_menuShow_triggered(QAction * act)
{
    InventoryProxyModel::ViewOption opt = 
        (InventoryProxyModel::ViewOption) act->data().toInt();
    proxyModelFolderTree->setViewOption(opt);
    proxyModelFileList->setViewOption(opt);
    
    // disable any previous action and check the new action entry
    QList<QWidget*> assocWidgets = act->associatedWidgets();
    I(assocWidgets.size() > 0);
    // we assume that this action is only assigned to one widget and that
    // is the QMenu widget we need to find out all the other actions
    QList<QAction*> list = assocWidgets[0]->actions();
    for (int i=0, j=list.size(); i<j; i++)
    {
        list.at(i)->setChecked(false);
    }
    act->setChecked(true);
}

void MainWindow::on_actionExpand_tree_triggered()
{
    bool expanded = actionExpand_tree->data().toBool();
    
    if (expanded)
        treeView->collapseAll();
    else
        treeView->expandAll();
    
    actionExpand_tree->setText(
        expanded ? tr("Expand tree") : tr("Collapse tree")
    );
    
    actionExpand_tree->setData(QVariant(!expanded));
}

void MainWindow::openRecentWorkspace()
{
    QAction *action = qobject_cast<QAction *>(sender());
    if (action)
    {
    	emit loadWorkspace(action->data().toString());
    }
}


void MainWindow::openRecentDatabase()
{
    QAction *action = qobject_cast<QAction *>(sender());
    if (action)
    {
    	emit loadDatabase(action->data().toString());
    }
}

void MainWindow::doUpdatePreviousWorkspacesMenu()
{
    // clear previous actions
    menuRecent_Workspaces->clear(); 
    
    QStringList previousWs = Settings::getItemList("RecentWorkspaceList");
    int elemCount = previousWs.size();
    if (elemCount == 0)
    {
        menuRecent_Workspaces->addAction(tr("No previous workspaces available."));
        return;
    }

    QAction *act;
    // add the new actions  
    for (int i = 0; i < elemCount; ++i)
    {
        act = menuRecent_Workspaces->addAction(
            tr("&%1 %2").arg(i + 1).arg(previousWs[i]),
            this,
            SLOT(openRecentWorkspace())
        );
        act->setData(previousWs[i]);
    }
}

void MainWindow::doUpdatePreviousDatabasesMenu()
{
    // clear previous actions
    menuRecent_Databases->clear();  
    
    QStringList previousDb = Settings::getItemList("RecentDatabaseList");
    int elemCount = previousDb.size();
    if (elemCount == 0)
    {
        menuRecent_Databases->addAction(tr("No previous databases available."));
        return;
    }

    QAction *act;
    // add the new actions  
    for (int i = 0; i < elemCount; ++i)
    {
        act = menuRecent_Databases->addAction(
            tr("&%1 %2").arg(i + 1).arg(previousDb[i]),
            this,
            SLOT(openRecentDatabase())
        );
        act->setData(previousDb[i]);
    }
}

void MainWindow::updateWindowList()
{
    // remove old actions
    QList<QAction*> actions = menuWindow->actions();
    for (int i=0, j=actions.size(); i<j; i++)
    {
        QAction * act = actions.at(i);
        if (act->data().isValid())
        {
            menuWindow->removeAction(act);
        }
    }
    
    // TODO: we should integrate a small indicator which window is actually
    // active here, but this needs central work in the App
    QList<MainWindow*> list = APP->windowList();
    for (int i=0, j=list.size(); i<j; i++)
	{
		QAction * act = menuWindow->addAction(
			tr("&%1 %2").arg(i + 1).arg(list.at(i)->windowTitle()),
            this,
			SLOT(activateOtherWindow())
		);
		act->setData(i);
    }
}

void MainWindow::activateOtherWindow()
{
    // FIXME: the list of windows is not constant, theoretically it might
    // be possible that we've just been notified to raise a window which
    // was just removed. Since we can't / won't transport pointers around
    // we try to identify the window by its index in the windowList again,
    // if this fails, we do nothing. It might occur that we raise the wrong
    // window by this, of course, but all this is better than the alternatives.
    QAction *action = qobject_cast<QAction *>(sender());
    if (action)
    {
        int idx = action->data().toInt();
        QList<MainWindow*> list = APP->windowList();
        if (list.size() > idx)
        {
            MainWindow * wnd = list.at(idx);
            wnd->activateWindow();
            wnd->raise();
        }
    }
}

void MainWindow::on_actionBring_all_to_front_triggered()
{
    // FIXME: we should implement something here which prevents window
    // flickering if all windows are already activated
    QList<MainWindow*> list = APP->windowList();
    for (int i=0, j=list.size(); i<j; i++)
    {
        MainWindow * wnd = list.at(i);
        wnd->activateWindow();
        wnd->raise();
    }
}

void MainWindow::on_actionKey_management_triggered()
{
    KeyManagement dialog(this);
    dialog.execDocumentModal();
}

void MainWindow:: on_actionChangeset_browser_triggered()
{
    ChangesetBrowser dialog(this);
    dialog.execDocumentModal();
}

void MainWindow::on_actionUpdate_workspace_triggered()
{
    UpdateWorkspace dialog(this);
    dialog.execDocumentModal();
}

void MainWindow::on_actionCommit_revision_triggered()
{
    CommitRevision dialog(this);
    dialog.execDocumentModal();
}

void MainWindow::on_actionCheckout_revision_triggered()
{
    CheckoutRevision dialog(this);
    int ret = dialog.execDocumentModal();
    
    if (ret == 0)
    {
        QDir checkoutDir = dialog.getSelectedCheckoutDirectory();
        QString revision = dialog.getSelectedRevision();
        WorkspaceCreator creator(this, revision, checkoutDir);
        
        if (!creator.run())
        {
            C("Could not create workspace.");
            return;
        }
        emit loadWorkspace(checkoutDir.absolutePath());
    }
}

// FIXME: the next two handlers should probably ensure that the according dialog
// is only opened once per application
void MainWindow::on_actionPreferences_triggered()
{
    Preferences dialog(this);
    dialog.exec();
}

void MainWindow::on_actionAbout_guitone_triggered()
{
    About dialog(this);
    dialog.exec();
}

void MainWindow::on_actionAbout_Qt_triggered()
{
    qApp->aboutQt();
}

void MainWindow::on_actionCheck_for_updates_triggered()
{
#ifdef Q_WS_MACX
    CocoaUtil::checkForUpdates();
#else
    ApplicationUpdate dlg(this);
    if (!dlg.updateAvailable())
    {
        QMessageBox::information(
            this,
            tr("No updates available"),
            tr("Your version of guitone (%1) is already up-to-date.")
                .arg(GUITONE_VERSION),
            QMessageBox::Ok
        );
        return;
    }
    dlg.exec();
#endif
}

void MainWindow::invalidWorkspaceFormat(const QString & error)
{
    QString workspacePath = MTN(this)->getNormalizedWorkspacePath();
    
    QMessageBox::critical(
        this,
        tr("Unable to load workspace"),
        tr("The workspace '%1' could not be loaded.\nmonotone returned:\n\n%2")
            .arg(workspacePath)
            .arg(error),
        QMessageBox::Ok
    );
    
    Settings::removeItemFromList("RecentWorkspaceList", workspacePath);
    
    emit updatePreviousWorkspacesMenu();
    
    close();
}

// FIXME: These two functions are actually a hack: we block the closing of this
// window if a non-blocking (non-modal) dialog is still open which expects that 
// its parent window pointer is still valid at some point. As dialogs can be 
// nested, a simple boolean value wouldn't be sufficient since the close action
// of the main window would get enabled again as soon as the first subsequent
// dialog has been closed, which would lead to segfaults for the one opened in
// first place. Of course the underlying bug (that the app segfaults at all)
// should be fixed someday...
void MainWindow::enableClosing()
{
    I(closeCounter > 0);
    closeCounter--;
}

void MainWindow::disableClosing()
{
    closeCounter++;
}

