/***************************************************************************
 *   Copyright (C) 2006 by Thomas Keller                                   *
 *   me@thomaskeller.biz                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "Preferences.h"
#include "Monotone.h"
#include "Settings.h"
#include "DebugLog.h"

#include <QFileDialog>
#include <QMessageBox>
#include <QHeaderView>

Preferences::Preferences(QWidget* parent) : Dialog(parent)
{
    setupUi(this);
    Dialog::init();
    
    mtnExecutablePath->setText(Settings::getMtnBinaryPath());
    
    enableConsoleLog->setCheckState(DebugLog::getConsoleLogEnabled() ?
                                    Qt::Checked : Qt::Unchecked);
    enableFileLog->setCheckState(DebugLog::getFileLogEnabled() ?
                                 Qt::Checked : Qt::Unchecked);
    
    enableFileLog->setText(enableFileLog->text().arg(DebugLog::logFilePath()));
    
    logLevel->addItem(tr("Low (fatal errors)"), 1);
    logLevel->addItem(tr("Medium (critical errors)"), 2);
    logLevel->addItem(tr("High (warnings)"), 3);
    logLevel->addItem(tr("All (debug messages)"), 4);
    
    logLevel->setCurrentIndex(DebugLog::getLogLevel()-1);
    
    checkForUpdates->setCheckState(
        Settings::getBool("CheckForUpdates", true) ?
        Qt::Checked : Qt::Unchecked
    );
    
    connect(
        selectMtnExecutable, SIGNAL(clicked()),
        this, SLOT(openFileBrowser())
    );
}

Preferences::~Preferences() {}

void Preferences::accept()
{
    QString oldPath = Settings::getMtnBinaryPath();
    QString newPath = mtnExecutablePath->text();
    
    // check if the new executable satisfies our version needs
    // we'll first check the program's plain version and if this
    // does not succeed we check for the interface version
    if (!Monotone::checkProgramVersion(newPath) &&
        !Monotone::checkInterfaceVersion(newPath))
    {
        QMessageBox::critical(
            this,
            tr("Error"),
            tr("The path to the monotone binary is either invalid "
             "or points to an older version of monotone. "
             "Guitone requires monotone version %1 "
             "or a monotone with interface version %2 or later.")
            .arg(Monotone::RequiredProgramVersion)
            .arg(Monotone::RequiredInterfaceVersion),
            QMessageBox::Ok, 0, 0
        );
        return;
    }
    
    
    if (oldPath != newPath)
    {
        QMessageBox::information(
             this,
             tr("Notice"),
             tr("You need to reload your current "
                "workspace or restart guitone to use "
                "the new monotone binary."),
             QMessageBox::Ok
        );
        
        Settings::setMtnBinaryPath(newPath);
    }
    
    bool consoleEnabled = enableConsoleLog->isChecked();
    bool fileEnabled = enableFileLog->isChecked();
    int level = logLevel->itemData(logLevel->currentIndex()).toInt();
    
    DebugLog::setConsoleLogEnabled(consoleEnabled);
    DebugLog::setFileLogEnabled(fileEnabled);
    DebugLog::setLogLevel(level);

    Settings::setBool("CheckForUpdates", checkForUpdates->isChecked());
    
    done(QDialog::Accepted);
}

void Preferences::openFileBrowser()
{
    QString file = QFileDialog::getOpenFileName(
         this,
         tr("Choose the monotone executable"),
         "",
         tr("Binaries (mtn mtn.exe)")
    );
    
    if (!file.isEmpty())
    {
        mtnExecutablePath->setText(file);
    }
}

