/***************************************************************************
 *   Copyright (C) 2007 by Thomas Keller                                   *
 *   me@thomaskeller.biz                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "RevisionManifest.h"
#include "Guitone.h"
#include "Platform.h"
#include "FileExporter.h"

#include <QHeaderView>
#include <QDir>
#include <QMessageBox>
#include <QMenu>
#include <QFileDialog>

RevisionManifest::RevisionManifest(QWidget* parent, QString rev)
    : Dialog(parent), revision(rev)
{
    setupUi(this);
    Dialog::init();
    
    setWindowTitle(windowTitle().arg(revision));
    
    manifestModel = new Manifest(this);
    manifestView->setModel(manifestModel);
    
    connect(
        saveSelected, SIGNAL(clicked()),
        this, SLOT(saveSelectedFiles())
    );
    
    connect(
        saveAll, SIGNAL(clicked()),
        this, SLOT(saveAllFiles())
    );
    
    connect(
        manifestView, SIGNAL(doubleClicked(const QModelIndex &)),
        this, SLOT(openFile(const QModelIndex &))
    );
    
    connect(
        manifestView, SIGNAL(contextMenuRequested(const QModelIndexList &, const QPoint &)),
        this, SLOT(contextMenuRequested(const QModelIndexList &, const QPoint &))
    );
        
    if (!manifestModel->readManifest(revision))
    {
        C("Couldn't execute readManifest");
        // FIXME: proper error handling here
    }
}

RevisionManifest::~RevisionManifest()
{
    delete manifestModel;
}

void RevisionManifest::openFile(const QModelIndex & index)
{
    if (!index.isValid()) return;
    
    ManifestEntry * entry = static_cast<ManifestEntry*>(index.internalPointer());
    
    // return silently
    if (entry->is_directory) return;
    
    Monotone * mtn = MTN(this);
    
    int commandNumber;
    QStringList cmd;
    cmd << "get_file" << entry->hash;

    if (!mtn->executeCommand(cmd, commandNumber) ||
        mtn->getReturnCode(commandNumber) > 0)
    {
        C("Cannot execute get_file");
        return;
    }
    
    //
    // FIXME: temporary files are currently saved into /tmp/guitone-<user>/<name>.ext
    // on unices and somewhere in %WINDIR%\temp\guitone-<user> or similar under Win32.
    // We need to check that
    //  a) user A can't read the temp files of user B (unsure in Win32...)
    //  b) what we do if two different instances of guitone run for the same
    //     user which both want the same filename
    //
    QString tempPath = QDir::tempPath();
    tempPath.append(QDir::separator()).append("guitone-").append(Platform::getUsername());
    QFileInfo path(tempPath);
    if (!path.isDir() && !QDir().mkpath(tempPath))
    {
        C("Could not create temporary guitone directory");
        return;
    }
    tempPath.append(QDir::separator()).append(entry->filename());
    
    QFile file(tempPath);
    if (file.exists()) file.remove();
    
    if (!file.open(QIODevice::WriteOnly))
    {
        C("Can't open temporary file for writing");
        return;
    }
    
    file.write(mtn->getRawData(commandNumber));
    file.close();
    
    if (!Platform::openFile(this, tempPath))
    {
        QMessageBox::critical(
            this,
            tr("Error"),
            tr("Unable to open files on your platform - please contact the "
            "author about this problem.")
        );
        
        file.remove();
        return;
    }
    
    QMessageBox::information(
        this,
        tr("Information"),
        tr("Please close this message to remove the temporary file.")
    );
    
    file.remove();
}

void RevisionManifest::contextMenuRequested(const QModelIndexList & indexList, const QPoint & pos)
{
    if (indexList.size() == 0) return;
    
    QMenu menu(this);
    QFont activeFont;
    activeFont.setBold(true);
    
    QAction * actOpenFile = menu.addAction(tr("Open file"));
    QAction * actSaveSelected = menu.addAction(tr("Save selected"));
    QAction * actSaveAll = menu.addAction(tr("Save all"));
    
    if (indexList.size() == 1 && 
        !static_cast<ManifestEntry*>(indexList.at(0).internalPointer())->is_directory)
    {
        actOpenFile->setFont(activeFont);
    }
    else
    {
        actOpenFile->setVisible(false);
    }
    
    QAction * act = menu.exec(pos);
    if (act == actOpenFile)
    {
        openFile(indexList.at(0));
    }
    else if (act == actSaveSelected)
    {
        saveSelectedFiles();
    }
    else if (act == actSaveAll)
    {
        saveAllFiles();
    }
}


void RevisionManifest::saveAllFiles()
{
    QModelIndexList list;
    int row = 0;
    while (true)
    {
        QModelIndex index(manifestModel->index(row++, 0, QModelIndex()));
        if (!index.isValid()) break;
        list.append(index);
    }
    
    QString exportDir = 
        QFileDialog::getExistingDirectory(0, tr("Select your export directory..."));
    if (!exportDir.isEmpty())
    {
        FileExporter exporter(this, revision, QDir(exportDir));
        
        FileEntryList entries = extractSelectedEntries(list);
        if (!exporter.run(entries))
        {
            QMessageBox::critical(
                this,
                tr("Error"),
                tr("The file export was aborted - please look in the log for more details.")
            );
            return;
        }
    }
}

void RevisionManifest::saveSelectedFiles()
{
    QModelIndexList indexes = manifestView->selectionModel()->selectedIndexes();
    FileEntryList entries = extractSelectedEntries(indexes);
    
    if (entries.size() == 0)
    {
        QMessageBox::information(
            this,
            tr("Information"),
            tr("Please select one or more files you want to save locally.")
        );
        return;
    }
    
    QString exportDir = 
        QFileDialog::getExistingDirectory(0, tr("Select your export directory..."));
    if (!exportDir.isEmpty())
    {
        FileExporter exporter(this, revision, QDir(exportDir));
        if (!exporter.run(entries))
        {
            QMessageBox::critical(
                this,
                tr("Error"),
                tr("The file export was aborted - please look in the log for more details.")
            );
            return;
        }
    }
}

FileEntryList RevisionManifest::extractSelectedEntries(const QModelIndexList & indexList)
{
    FileEntryList entries;
    for (int i=0, j=indexList.size(); i<j; i++)
    {
        ManifestEntry * entry = 
            static_cast<ManifestEntry*>(indexList.at(i).internalPointer());
        
        if (entry->is_directory)
        {
            entries.append(FileEntry(entry->path, true));
            entries += addEntriesRecursive(entry);
        }
        else
        {
            entries.append(FileEntry(entry->path, false));
        }
    }
    
    return entries;
}

FileEntryList RevisionManifest::addEntriesRecursive(ManifestEntry * entry)
{
    FileEntryList entries;
    for (int i=0, j=entry->childCount(); i<j; i++)
    {
        if (entry->children.at(i)->is_directory)
        {
            entries.append(FileEntry(entry->children.at(i)->path, true));
            entries += addEntriesRecursive(entry->children.at(i));
        }
        else
        {
            entries.append(FileEntry(entry->children.at(i)->path, false));
        }
    }
    return entries;
}

